/******************************************************************************
 *
 * NAME
 *    decode.h
 *   
 * DATE
 *    1.8.1998
 *
 * IMPLEMENTATION 
 *    Bertram Ganz (ganz@informatik.uni-freiburg.de)
 *    Martin Mueller (mueller@www.fr.bw.schule.de)
 *    
 *    The fractal part is based on Yuval Fishers dec.c code 
 *   
 * MODIFICATIONS
 *   
 * DESCRIPTION
 *    Decode an image encoded with a quadtree Fractal-MRSGVQ hybrid scheme
 *
 * USAGE
 *    dec -[options] [inputfile [outputfile]]
 * 
 * OPTIONS 
 *       (# = number), defaults show in ()
 *    -b VQ Blocks are decoded white. (off)
 *    -B Fractal Blocks are decoded black. (off)
 *    -f # scale factor of output size. (1.000000)
 *    -i file name. An initial image to iteration from.
 *    -n # no. of decoding iterations. (10)
 *    -N # maximum allowed scaling. (1.000000)
 *    -V4 # maximum scaling in VQ encoding (4x4). (250.0)
 *    -V8 # maximum scaling in VQ encoding (8x8). (500.0)
 *    -V9 # maximum scaling in VQ encoding (16x16). (1000.0)
 *    -s # number of scaling quantizing bits. (5)
 *    -o # number of offset quantizing bits. (7)
 *    -P   output the partition into the output file. (off)
 *    -p   supress artifact postprocessing. (off)
 *
 *****************************************************************************/

/* The following belong in a dec.h file, but nevermind...              */
/* -----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/stat.h>

#define LEAF        1
#define NODE        0
#define CLUSTERS    4096
#define DEBUG       0
#define GREY_LEVELS 255
#define IMAGE_TYPE unsigned char /* may be different in some applications */

/* The following #define allocates an hsize x vsize  matrix of type type */
#define matrix_allocate(matrix, hsize,vsize, TYPE) {\
    TYPE *imptr; \
    int _i; \
    matrix = (TYPE **)malloc(vsize*sizeof(TYPE *));\
    imptr = (TYPE*)malloc((long)hsize*(long)vsize*sizeof(TYPE));\
    if (imptr == NULL) \
        fatal("\nNo memory in matrix allocate."); \
    for (_i = 0; _i<vsize; ++_i, imptr += hsize) \
        matrix[_i] = imptr; \
 }

#define bound(a)   ((a) < 0.0 ? 0 : ((a)>255.0? 255 : a))

/* various function declarations to keep compiler warnings away           */
void fatal();

/* ---------------------------------------------------------------------- */
/* the following is the HUFFMAN CODE used for decoding the
   OMSGVQ scaling factors */
const char huff_codeword[3][32][13] =  /* 16x16 */ 
                                {"0011011010","1101010101","1101010100", 
				 "0011011000","10100100","1101011","1010001",
				 "1010000","101011","101010","11011","1011",
				 "1001","111","011","10000","10001","010",
				 "000","0010","1100","00111","001100","110100",
				 "1010011","0011010","00110111","10100101",
				 "11010100","110101011","0011011011",
				 "0011011001",
				 /* 8x8 */
				 "10011110","10011111","01101100","00110110", 
				 "00110101","0001101","011010","001100","10010",
				 "00111","1010","1111","1011","1110","0101",
				 "10001","1001110","0100","110","0000","0010",
				 "0111","00010","01100","10000","000111",
				 "100110","0001100","0110111","00110100",
				 "00110111","01101101",
				 /* 4x4 */  
				 "111000","00011111","1111111","0011101", 
				 "111110","111010","000001","11110","00001",
				 "00101","00010","1100","1010","0100","0111",
				 "001111","1111110","0110","0101","1001",
				 "1011","1000","1101","00100","00110","000000",
				 "000110","111011","0001110","0011100",
				 "00011110","111001"};

IMAGE_TYPE **image,*imptr,**image1; 
                             /* The input image data and a dummy          */ 

double max_scale = 1.0,      /* maximum allowable grey level scale factor */
       max_scale_vq[3];

int    vec_bits = 8,         /* Bits for cluster position                 */
       s_bits = 5,           /* number of bits used to store scale factor */
       o_bits = 7,           /* number of bits used to store offset       */
       hsize = -1,           /* The horizontal size of the input image    */
       vsize = -1,           /* The vertical size                         */
       scaledhsize,          /* hsize*scalefactor                         */
       scaledvsize,          /* vsize*scalefactor                         */
       size,                 /* largest square image that fits in image   */
       dim=16,               /* dimension of cluster center vectors       */
       min_part = 3,         /* min and max _part determine a range of    */
       max_part = 4,         /* Range sizes from hsize>>min to hsize>>max */
       dom_step = 4,         /* Density of domains relative to size       */
       dom_step_type = 0,    /* Flag for dom_step a multiplier or divisor */
       dom_type = 0,         /* Method of generating domain pool 0,1,2..  */
       post_process = 1,     /* Flag for postprocessing.                  */
       num_iterations= 10,   /* Number of decoding iterations used.       */
       *no_h_domains,        /* Number of horizontal domains.             */
       *domain_hstep,        /* Domain density step size.                 */
       *domain_vstep,        /* Domain density step size.                 */
       *bits_needed,         /* Number of bits to encode domain position. */
       output_partition=0,   /* A flag for outputing the partition        */
       max_exponent,         /* The max power of 2 side of square image   */
                             /* that fits in our input image.             */
       print_vq_part=0,      /* A flag for outputing the OMSGVQ         */
                             /* partition.                                */
       print_frac_part=0;    /* A flag for outputing the partition of     */
                             /* fractally coded image blocks.             */

struct transformation_node {
       int    rx,ry,         /* The range position and size in a trans.   */
              xsize, ysize,  
              rrx,rry,
              dx,dy;         /* The domain position.                      */
       int sym_op;           /* The symmetry operation used in the trans. */
       int depth;            /* The depth in the quadtree partition.      */
       double scale, offset; /* scaling and offset values.                */
       struct transformation_node *next; /* The next trans. in list       */
} transformations, *trans; 

struct transformation_node_VQ {
       int    rx,ry,         /* The range position and size in a trans.   */
              xsize, ysize,  
              rrx,rry,
              vec_pos;       /* The Vector position.                      */
       int sym_op;           /* The symmetry operation used in the trans. */
       int depth;            /* The depth in the quadtree partition.      */
       double scale, offset; /* scaling and offset values.               */
       struct transformation_node_VQ *next;    /* The next trans. in list */
} transformations_VQ, *trans_VQ; 

struct huffman_tree_node {   /* huff tree for decoding VQ scal. factors  */
    int index; /* scal. factor index in final leaf */
    int type;  /* inner node or final leaf */
    struct huffman_tree_node *left, *right;
} **huff_tree;

struct counter_data {  /* counter data */
    int *vq,     /* each quadtree layer has its own countrer */
        *fr,
        oqueue,  /* counter for queue containing offset information */
        mqueue;  /* counter for queue containing method information */
} counter;

FILE *input,*output,*other_input; 
    
double ***centers,    /* List of Cluster centers */
       **centers_64,
       **centers_256;

int    *method_queue,
       *offset_queue;

/* FATAL is for when there is a fatal error... print a message and exit */
void fatal(s)
char *s;
{
     printf("\n%s\n",s);
     exit(-1);
}

/**************************************************************/
/* UNPACK VALUE using size bits read from fin.                */
/**************************************************************/
long unpack(int size, int which_file, FILE *fin)
{
    int i;
    int value = 0;
    static int ptr[3] = {1,1,1}; /* how many bits are packed in sum so far */
    static int sum[3]; 


    /* size == -2 means we initialize things */
    if (size == -2) {
        sum[which_file] = fgetc(fin);
        sum[which_file] <<= 1;
        return((long)0);
    }

    /* size == -1 means we want to peek at the next bit without */
    /* advancing the pointer */
    if (size == -1)
        return((long)((sum[which_file]&256)>>8));

     for (i=0; i<size; ++i, ++ptr[which_file],  sum[which_file] <<= 1) {
        if (sum[which_file] & 256) value |= 1<<i;

         if (ptr[which_file] == 8) {
            sum[which_file] = getc(fin);
            ptr[which_file]=0;
        }
     }
    return((long)value);
}

/********************************/
/* NORM of a                    */
/********************************/
double norm(double *a, int dim)  
{
  int i;
  double s=0.;
  for(i=0; i<dim; i++)
    s+=a[i]*a[i];
  return s;
}


/*************   M A I N  *******************/
/*******/
/*******/
int main(int argc, char **argv)
{
    /* Defaults are set initially */
    double         scalefactor = 1.0;           /* Scale factor for output */
    char           inputfilename[200];
    char           outputfilename[200];
    char           other_input_file[200];
    double         a, b;
    double         m=0, rms, PSNR;
    int            i,j, x_exponent, y_exponent;
    int            domain_size, no_domains;
    int            depth;
    int            k;
    double         vecnorm;
    FILE           *cluster_file;
    char           ch[200];

    inputfilename[0] = 1;    /* We initially set the input to this and */
    outputfilename[0] = 1;   /* then check if the input/output names   */
    other_input_file[0] = 1; /* have been set below.                   */

    max_scale_vq[0] = 250;
    max_scale_vq[1] = 500;
    max_scale_vq[2] = 1000;


    /* SCAN through the INPUT LINE and READ in the ARGUMENTS */
    for (i=1; i<argc; ++i) 
       if (argv[i][0] != '-' ) 
            if (inputfilename[0] == 1)
                strcpy(inputfilename, argv[i]);
            else if (outputfilename[0] == 1)
                strcpy(outputfilename, argv[i]);
            else;
       else { /* we have a flag */
            if (strlen(argv[i]) == 1) break;
            switch(argv[i][1]) {
	    case 'b': print_vq_part = 1;
	      break;
	    case 'B': print_frac_part = 1;
	      break;
	    case 'i': strcpy(other_input_file,argv[++i]); 
	      break;
	    case 'n': num_iterations = atoi(argv[++i]);
	      break;
	    case 'f': scalefactor = atof(argv[++i]);
	      break;
	    case 'P': output_partition = 1;
	      break;
	    case 'p': post_process = 0;
	      break;
	    case 's': s_bits = atoi(argv[++i]);
	      break;
	    case 'o': o_bits = atoi(argv[++i]);
	      break;
	    case 'N': max_scale = atof(argv[++i]);
                         break;
	    case 'V': if (strlen(argv[i]) == 2) fatal("Falscher Parameter");
	      switch(argv[i][2]) 
		{
		case '4': max_scale_vq[0] = atof(argv[++i]);
		  break;
		case '8': max_scale_vq[1] = atof(argv[++i]);
		  break;
		case '9': max_scale_vq[2] = atof(argv[++i]);
		  break;
		default: break;
		}; break;
	    case '?':
	    case 'H':
	    default:
	      printf("\nUsage: dec -[options] [inputfile [outputfile]]");
	      printf("\nOptions are: (# = number), defaults show in ()");
	      printf("\n  -b VQ Blocks are decoded white. (off)");
	      printf("\n  -B Fractal Blocks are decoded black. (off)");
	      printf("\n  -f # scale factor of output size. (%lf)", scalefactor);
	      printf("\n  -i file name. An initial image to iteration from.");
	      printf("\n  -n # no. of decoding iterations. (%d)", num_iterations);
	      printf("\n  -N # maximum allowed scaling. (%lf)", max_scale);
	      printf("\n  -V4 # maximum scaling in VQ encoding (4x4). (%f)",
		     max_scale_vq[0]);
	      printf("\n  -V8 # maximum scaling in VQ encoding (8x8). (%f)",
		     max_scale_vq[1]);
	      printf("\n  -V9 # maximum scaling in VQ encoding (16x16). (%f)",
		     max_scale_vq[2]);
	      printf("\n  -s # number of scaling quantizing bits. (%d)",s_bits);
	      printf("\n  -o # number of offset quantizing bits. (%d)",o_bits);
	      printf("\n  -P   output the partition into the output file. (off)");
	      printf("\n  -p   supress artifact postprocessing. (off)");
	      fatal("\n        --------- Version  1.0 --------------");
	    }
       }

    if (inputfilename[0] == 1) strcpy(inputfilename, "lenna.trn");
    if (outputfilename[0] == 1) strcpy(outputfilename, "lenna.out");

    if ((input = fopen(inputfilename, "r")) == NULL)
        fatal("Can't open input file.");
        
    unpack(-2,2,input); /* INITIALIZE the UNPACKING routine */

    /* READ THE HEADER DATA from the input file. This should probably */
    /* be put into one read which reads a structure with the info     */
    min_part = (int)unpack(4,2,input);
    max_part = (int)unpack(4,2,input);
    dom_step = (int)unpack(4,2,input);
    dom_step_type = (int)unpack(1,2,input);
    dom_type = (int)unpack(2,2,input);
    hsize = (int)unpack(12,2,input);
    vsize = (int)unpack(12,2,input);
    vec_bits = (int)unpack(4,2,input);

    /* READ THE METHOD AND OFFSET DATA from inputfile */
    read_queues();

    /* we now COMPUTE SIZE */
    x_exponent = (int)floor(log((double)hsize)/log(2.0));
    y_exponent = (int)floor(log((double)vsize)/log(2.0));
   
    /* exponent is min of x_ and y_ exponent */
    max_exponent = (x_exponent > y_exponent ? y_exponent : x_exponent);
    /* size is the size of the largest square that fits in the image */
    /* It is used to compute the domain and range sizes.             */
    size =  1<<max_exponent; 

    /* ALLOCATE MEMORY FOR THE OUTPUT IMAGE. Allocating one chunck saves  */
    /* work and time later.                                              */
    scaledhsize = (int)(scalefactor*hsize);
    scaledvsize = (int)(scalefactor*vsize);
    matrix_allocate(image, scaledhsize,scaledvsize, IMAGE_TYPE);
    matrix_allocate(image1, scaledhsize, scaledvsize, IMAGE_TYPE);

    /* ALLOCATE MEMORY for huffmantrees. They are used for variable length-
       decoding of VQ scaling factors */
    depth=max_part-min_part+1;    
    huff_tree = (struct huffman_tree_node **)
      malloc(3*sizeof(struct huffman_tree_node *));
    for (i=0; i<3; ++i)
      { 
	huff_tree[i]=(struct huffman_tree_node *)
	             malloc(sizeof(struct huffman_tree_node));
	huff_tree[i]->type=LEAF;
      }
	
    /* BUILD HUFFMAN TREES by reading huffman codewords */
    for (i=0; i<3; ++i) 
      for (j=0; j<32; j++)
	build_huffman_tree(huff_tree[i], huff_codeword[i][j], j, 0);

    /* Counter ALLOCATION */
    counter.vq = (int *) malloc(depth * sizeof(int));
    counter.fr = (int *) malloc(depth * sizeof(int));
    
    for(i=0; i<depth; i++) /* INITIALIZATION of counter data */
      {
	counter.vq[i] = 0;
	counter.fr[i] = 0;
      }

    if (other_input_file[0] != 1) 
      {
        other_input = fopen(other_input_file, "r");
        i = fread(image[0], sizeof(IMAGE_TYPE), 
				scaledhsize*scaledvsize, other_input);
        if (i < scaledhsize*scaledvsize) 
           fatal("Couldn't read input... not enough data.");
        else
            printf("\n%d pixels read from %s.\n", i,other_input_file);
        fclose(other_input);
      }       
  
    /* READ-IN cluster center vectors */
    centers  = (double ***) malloc(depth * sizeof(double**)); 
    if ((cluster_file = fopen("cen_16.som", "r")) == NULL)
      fatal("Can't open center file.");
    for (i=0; i < depth; ++i) 
      {
	fgets(ch , 200, cluster_file); /* SKIP first line */
	centers[i]  = (double **) malloc(CLUSTERS * sizeof(double*)); 
	
	/* READ-IN cluster center vectors */
	for(j=0; j<CLUSTERS; j++)  
	  {
	    centers[i][j] = (double *) malloc(dim * sizeof(double)); 
	    for(k=0; k<dim; k++)
	      if (fscanf(cluster_file, "%lf", &centers[i][j][k])==EOF) goto ende;
	    /* break loop at end of file */    
	    
	    /* NORMALIZE cluster center vectors */
	    vecnorm = sqrt(norm(centers[i][j], dim));
	    for(k=0; k<dim; k++)
	      centers[i][j][k] = centers[i][j][k]/vecnorm;
	  }
        ende:
	rewind(cluster_file);
      }
    fclose(cluster_file);
    
    /* ************************************************************ */
    if ((cluster_file = fopen("cen_64.som", "r")) == NULL)
      fatal("Can't open center file.");

    fgets(ch , 200, cluster_file);    /* SKIP first line */  
    centers_64  = (double **) malloc(CLUSTERS * sizeof(double*)); 

    /* READ IN cluster center vectors */
    for(j=0; j<CLUSTERS; j++)  
      {
	centers_64[j] = (double *) malloc(64 * sizeof(double)); 
	for(k=0; k<64; k++)
	  if (fscanf(cluster_file, "%lf", &centers_64[j][k])==EOF) goto ende_2;
	/* break loop at end of file */    
	
	/* NORMALIZE cluster center vectors */
	vecnorm = sqrt(norm(centers_64[j], 64));
	for(k=0; k<64; k++)
	  centers_64[j][k] = centers_64[j][k]/vecnorm;
      }
    ende_2:
    rewind(cluster_file);
    fclose(cluster_file);

    /* ************************************************************ */
    if ((cluster_file = fopen("cen_256.som", "r")) == NULL)
      fatal("Can't open center file.");

    /* INTIALIZATION of cluster list */
    fgets(ch , 200, cluster_file);  /* SKIP first line */  
    centers_256  = (double **) malloc(CLUSTERS * sizeof(double*)); 
    
    /* READ Cluster Center Vectors*/
    for(j=0; j<CLUSTERS; j++)  
      {
	centers_256[j] = (double *) malloc(256 * sizeof(double)); 
	for(k=0; k<256; k++)
	  if (fscanf(cluster_file, "%lf", &centers_256[j][k])==EOF) goto ende_3;
	/* BREAK loop at end of file */    
	
	/* NORMALIZE Cluster Center Vectors */
	vecnorm = sqrt(norm(centers_256[j], 256));
	for(k=0; k<256; k++)
	  centers_256[j][k] = centers_256[j][k]/vecnorm;
      }
    ende_3:
    rewind(cluster_file);
    fclose(cluster_file);
    /* ************************************************************ */

    /* since max_ and min_ part are variable, these must be allocated */
    bits_needed = (int *)malloc(sizeof(int)*depth);
    no_h_domains = (int *)malloc(sizeof(int)*depth);
    domain_hstep = (int *)malloc(sizeof(int)*depth);
    domain_vstep = (int *)malloc(sizeof(int)*depth);
    
    /* compute bits needed to read each domain type */
    for (i=0; i < depth; ++i) {
       /* first compute how many domains there are horizontally */
       domain_size = size >> (min_part+i-1);
       if (dom_type == 2) 
             domain_hstep[i] = dom_step; 
       else if (dom_type == 1)
             if (dom_step_type ==1)
                domain_hstep[i] = (size >> (max_part- i-1))*dom_step;
             else 
                domain_hstep[i] = (size >> (max_part - i-1))/dom_step;
       else 
             if (dom_step_type ==1)
                domain_hstep[i] = domain_size*dom_step;
             else 
                domain_hstep[i] = domain_size/dom_step;

       no_h_domains[i] = 1+(hsize-domain_size)/domain_hstep[i];

       /* now COMPUTE HOW MANY DOMAINS there are VERTICALLY */
       if (dom_type == 2) 
             domain_vstep[i] = dom_step; 
       else if (dom_type == 1)
             if (dom_step_type ==1)
             domain_vstep[i] = (size >> (max_part - i-1))*dom_step;
             else 
             domain_vstep[i] = (size >> (max_part - i-1))/dom_step;
       else 
             if (dom_step_type ==1)
                domain_vstep[i] = domain_size*dom_step;
             else 
                domain_vstep[i] = domain_size/dom_step;

       no_domains = 1+(vsize-domain_size)/domain_vstep[i];
       bits_needed[i] = ceil(log((double)no_domains*(double)no_h_domains[i])/
                           log(2.0)); 

       /* max_scale_vq */
       if ((max_part-min_part == 0) && (domain_size == 8))
	 max_scale_vq[1] = max_scale_vq[0];
       if ((max_part-min_part == 0) && (domain_size == 16))
	 max_scale_vq[0] = max_scale_vq[1];
       if ((max_part-min_part == 0) && (domain_size == 32))
	 max_scale_vq[0] = max_scale_vq[2];
       if ((max_part-min_part != 0) && (domain_size == 16) &&
	   (i == max_part-min_part) )
	 max_scale_vq[0] = max_scale_vq[1];      
     }

    if ((output = fopen(outputfilename, "w")) == NULL)
            fatal("Can't open output file.");

    /* READ IN the TRANSFORMATION DATA */
    trans = &transformations;
    trans_VQ = &transformations_VQ;
    printf("\nReading transformations.....");
    fflush(stdout);

    partition_image(0, 0, hsize,vsize );

    fclose(input);
    printf("Done.\n");
    fflush(stdout);
    if (scalefactor != 1.0) {
      printf("\nScaling image to %d x %d.", scaledhsize,scaledvsize);
      scale_transformations(scalefactor); 
    }
 
    /* when we OUTPUT THE PARTITION, we just read the transformations */
    /* in and write them to the outputfile                            */ 
    if (output_partition) {
          fprintf(output,"\n%d %d\n %d %d\n%d %d\n\n", 
                  0, 0, scaledhsize, 0, scaledhsize, scaledvsize);
          printf("\nOutputed partition data in %s\n",outputfilename);
          fclose(output);
          return;
    }

    /* INITIALIZATION */
    for(i=0; i<depth; i++) 
      printf("\nTransformations(fractal)\[%i\] = %i \t Transformations(VQ)\[%i\] = %i",i, counter.fr[i], i, counter.vq[i]);
    fflush(stdout);
    
    for (i=1; i <depth; ++i) 
      {
	counter.vq[0]+=counter.vq[i];
	counter.fr[0]+=counter.fr[i];
      }
    printf("\n\nNumber of VQ transf.      = %i", counter.vq[0]);
    printf("\nNumber of fractal transf. = %i\n", counter.fr[0]);

    image = image1;

    /* now we know all relevant data. START OMSGVQ part 
       and iterations (fractal)*/
    apply_transformations_VQ();
    if(!print_frac_part)
      for (i=0; i<num_iterations; ++i) 
	apply_transformations();
      
    if (post_process) 	
      smooth_image();

    i = fwrite(image[0], sizeof(IMAGE_TYPE), scaledhsize*scaledvsize, output);
    if (i < scaledhsize*scaledvsize) 
        fatal("Couldn't write output... not enough disk space ?.");
    else
        printf("\n%d pixels written to output file.\n", i);

    fclose(output);
    
    /* FREE allocated memory for HUFFMANTREES */
    for (i=0; i<3; ++i) 
      free_huffman_tree(huff_tree[i]);

    return(0);    
}

/*****************************************************************/
/* READ_QUEUES reads the method and offset data from inputfile. */
/* First we have to check wether gzip was used by the encoder    */
/* or not. Two queues are filled with the data for access in     */
/* read_maindata later (first-in first-out principle).           */
/*****************************************************************/
read_queues()
{
  int methodfile_bytes, offsetfile_bytes;
  int gzip;
  char *data;
  int i;
  struct stat file_data; 
  FILE *file;

  /* READ IN information about used method */
  gzip = (int)unpack(1,2,input);
  methodfile_bytes = (int)unpack(12,2,input);
  if(gzip){  /* gzip was used, so CALL GUNZIP */
    data = (char *)malloc(sizeof(char)*methodfile_bytes);
    fseek(input,-1,SEEK_CUR);
    fread(data,sizeof(char),methodfile_bytes,input);
    if ((file = fopen("methodfile.gz", "w")) == NULL)
      fatal("Can't open methodfile.gz");
    fwrite(data,sizeof(char),methodfile_bytes,file);
    fclose(file);
    free(data);
    system("gzip -d -9 -f methodfile.gz > /dev/null");
    if ((file = fopen("methodfile", "r")) == NULL)
      fatal("Can't open methodfile");
    stat("methodfile", &file_data);
    unpack(-2,0,file);
    method_queue = (int *)malloc(sizeof(int)*file_data.st_size*8);
    for(i=0;i<file_data.st_size*8;i++)
      method_queue[i] = unpack(1,0,file);
    fclose(file);
    system("rm methodfile");
    unpack(-2,2,input);
  }
  else { /* READ-IN method information directly */ 
    method_queue = (int *)malloc(sizeof(int)*methodfile_bytes*8);
    for(i=0;i<methodfile_bytes*8;i++)
      method_queue[i] = unpack(1,2,input);
  }
  
  /* READ IN information about offsets */
  gzip = (int)unpack(1,2,input);
  offsetfile_bytes = (int)unpack(15,2,input);
  if(gzip){  /* gzip was used, so CALL GUNZIP */
    data = (char *)malloc(sizeof(char)*offsetfile_bytes);
    fseek(input,-1,SEEK_CUR);
    fread(data,sizeof(char),offsetfile_bytes,input);
    if ((file = fopen("offsetfile.gz", "w")) == NULL)
      fatal("Can't open offsetfile.gz");
    fwrite(data,sizeof(char),offsetfile_bytes,file);
    fclose(file);
    free(data);
    system("gzip -d -9 -f offsetfile.gz > /dev/null");
    if ((file = fopen("offsetfile", "r")) == NULL)
      fatal("Can't open offsetfile");
    stat("offsetfile", &file_data);
    unpack(-2,1,file);
    offset_queue = (int *)malloc(sizeof(int)*file_data.st_size);
    for(i=0;i<file_data.st_size;i++)
      offset_queue[i] = unpack(8,1,file);
    fclose(file);
    system("rm offsetfile");
    unpack(-2,2,input);
  }
  else { /* READ-IN offset information directly */ 
    offset_queue = (int *)malloc(sizeof(int)*offsetfile_bytes);
    for(i=0;i<offsetfile_bytes;i++)
      offset_queue[i] = unpack(8,2,input);
  }
}

/****************************************************************/
/* READ IN THE TRANSFORMATION DATA from *input.                 */
/* This is a recursive routine whose recursion tree follows the */
/* recursion done by the encoding program.                      */
/****************************************************************/
read_maindata(int atx, int aty, int xsize, int ysize, int depth)
{
    /* Having all these locals in a recursive procedure is hard on the  */
    /* stack.. but it is more readable.                                 */
    int i,
        sym_op,                   /* A symmetry operation of the square */
        ialpha,                   /* Intgerized scalling factor         */
        ibeta;                    /* Intgerized offset                  */

    long domain_ref; 

    double alpha, beta;

    /* keep breaking it down until we are small enough */
    if (depth < min_part) {
         read_maindata(atx,aty, xsize/2, ysize/2, depth+1);
         read_maindata(atx+xsize/2,aty, xsize/2, ysize/2, depth+1);
         read_maindata(atx,aty+ysize/2,xsize/2, ysize/2,  depth+1);
         read_maindata(atx+xsize/2,aty+ysize/2,xsize/2,ysize/2,depth+1);
         return;
    }

    if (depth < max_part && unpack(1,2,input)) {
         /* A 1 means we subdivided.. so QUADTREE down */
         read_maindata(atx,aty, xsize/2, ysize/2, depth+1);
         read_maindata(atx+xsize/2,aty, xsize/2, ysize/2, depth+1);
         read_maindata(atx,aty+ysize/2, xsize/2, ysize/2, depth+1);
         read_maindata(atx+xsize/2,aty+ysize/2,xsize/2,ysize/2,depth+1);
    }
    else 
      {
	if ( method_queue[counter.mqueue++] == 1)
	  { 
	    /* we have a to READ OMSGVQ data*/
	    counter.vq[depth-min_part]++;
	    /* read in variable length code */
	    i=(xsize == 16) ? 0 : ( (xsize == 8) ? 1 : 2 ); 
	    read_sbits(huff_tree[i], &ialpha);
	    ibeta = offset_queue[counter.oqueue++];
	    trans_VQ->next = (struct transformation_node_VQ *)
	      malloc(sizeof(struct transformation_node_VQ ));
	    trans_VQ = trans_VQ->next;

	    trans_VQ->next=NULL; /* not automatically set by MALLOC */ 
	    alpha =(double)ialpha/(double)((1<<s_bits)-1)*2.0*
	      max_scale_vq[max_part-depth]-max_scale_vq[max_part-depth];  
	    beta = (double)ibeta/(double)((1<<o_bits)-1)*GREY_LEVELS;
	    trans_VQ->scale = alpha;
	    trans_VQ->offset = beta;
	    trans_VQ-> sym_op  = (int)unpack(3,2, input);
	    trans_VQ-> vec_pos = (int)unpack(vec_bits,2, input);
	    trans_VQ->rx = atx;
	    trans_VQ->ry = aty;
	    trans_VQ->depth = depth;
	    trans_VQ->xsize = xsize;
	    trans_VQ->ysize = ysize;
	    trans_VQ->rrx = atx + xsize;
	    trans_VQ->rry = aty + ysize;
	    
	  }	
	else if (method_queue[counter.mqueue++] == 0)
	  { /* we have a VQ transformation to read with s=0 */    
	    counter.vq[depth-min_part]++;
	    ibeta = offset_queue[counter.oqueue++];
	    trans_VQ->next = (struct transformation_node_VQ *)
	      malloc(sizeof(struct transformation_node_VQ ));
	    trans_VQ = trans_VQ->next;
	    
	    trans_VQ->next=NULL;  /* not automatically set by MALLOC */ 
	    trans_VQ->scale = 0.;
	    beta  = (double)ibeta/(double)((1<<o_bits)-1)*GREY_LEVELS;
	    trans_VQ->offset = beta;
	    trans_VQ->sym_op  = 0;
	    trans_VQ->vec_pos = 0;
	    trans_VQ->rx = atx;
	    trans_VQ->ry = aty;
	    trans_VQ->depth = depth;
	    trans_VQ->xsize = xsize;
	    trans_VQ->ysize = ysize;
	    trans_VQ->rrx = atx + xsize;
	    trans_VQ->rry = aty + ysize;
	  }
	else
	  { /* use fractal */
	    counter.fr[depth-min_part]++;
	    ialpha = (int)unpack(s_bits,2, input);     
	    ibeta = offset_queue[counter.oqueue++];
	    trans->next = (struct transformation_node *)
	      malloc(sizeof(struct transformation_node ));
	    trans = trans->next;
	    
	    trans->next=NULL; /* not automatically set by MALLOC */ 
	    alpha = (double)ialpha/(double)((1<<s_bits)-1)*
	            (2.0*max_scale)-max_scale; 
	    beta = (double)ibeta/(double)((1<<o_bits)-1)*
	      (GREY_LEVELS);
	    trans->scale = alpha;
	    trans->offset = beta;
	    trans->sym_op = (int)unpack(3,2,input);
	    domain_ref = unpack(bits_needed[depth-min_part],2,input);
	    trans->dx = (double)(domain_ref % no_h_domains[depth-min_part])
	      * domain_hstep[depth-min_part];
	    trans->dy = (double)(domain_ref / no_h_domains[depth-min_part])
	      * domain_vstep[depth-min_part];
	    trans->rx = atx;
	    trans->ry = aty;
	    trans->depth = depth;
	    trans->rrx = atx + xsize;
	    trans->rry = aty + ysize;
	  }
	
	if (output_partition) 
	  fprintf(output,"\n%d %d\n %d %d\n%d %d\n\n", 
                  atx,       vsize-aty-ysize, 
                  atx,       vsize-aty, 
		  atx+xsize, vsize-aty);
      }
}
     
/**************************************************************/
/* APPLY the TRANSFORMATIONS once for the VQ part             */
/**************************************************************/
apply_transformations_VQ()
{
    int	   i,j,i1,j1,k,l,count=0;
    double pixel, *center_point;


    trans_VQ = &transformations_VQ;
    while (trans_VQ->next != NULL) {
      trans_VQ = trans_VQ->next;
      ++count;

      l = trans_VQ->xsize >> 3;  /*adaption to variable sizes */

      if (trans_VQ->xsize == 4)
	center_point = centers[0][trans_VQ->vec_pos];
      else if (trans_VQ->xsize == 8)
	center_point = centers_64[trans_VQ->vec_pos];
      else
	center_point = centers_256[trans_VQ->vec_pos];
	

#define COMPUTE_LOOP_VQ              {                                      \
	k =  (i1 << (2+l))+j1;                                        \
        pixel = center_point[k];                           \
	if(!print_vq_part)                                                  \
          image[j][i] = bound(0.5 + trans_VQ->scale*pixel+trans_VQ->offset);\
	else                                                                \
    	  image[j][i] = GREY_LEVELS;                                        \
	  image1[j][i] = image[j][i];                                         \
        }

        switch(trans_VQ->sym_op) {
           case 0: for (i=trans_VQ->rx, i1 = 0; i<trans_VQ->rrx; ++i, ++i1)
                   for (j=trans_VQ->ry, j1=0; j<trans_VQ->rry; ++j, ++j1) 
                       COMPUTE_LOOP_VQ
                   break;
           case 1: for (j=trans_VQ->ry, i1 = 0; j<trans_VQ->rry; ++j, ++i1)
                   for (i=trans_VQ->rrx-1, j1=0; i>=(int)trans_VQ->rx; --i, ++j1) 
                       COMPUTE_LOOP_VQ
                   break;
           case 2: for (i=trans_VQ->rrx-1,i1 = 0; i>=(int)trans_VQ->rx; --i, ++i1)
                   for (j=trans_VQ->rry-1,j1=0; j>=(int)trans_VQ->ry; --j, ++j1) 
                       COMPUTE_LOOP_VQ
                   break;
           case 3: for (j=trans_VQ->rry-1,i1 = 0; j>=(int)trans_VQ->ry; --j, ++i1)
                   for (i=trans_VQ->rx, j1=0; i<trans_VQ->rrx; ++i, ++j1) 
                       COMPUTE_LOOP_VQ
                   break;
           case 4: for (j=trans_VQ->ry, i1 = 0; j<trans_VQ->rry; ++j, ++i1)
                   for (i=trans_VQ->rx, j1=0; i<trans_VQ->rrx; ++i, ++j1) 
                       COMPUTE_LOOP_VQ
                   break;
           case 5: for (i=trans_VQ->rx, i1 = 0; i<trans_VQ->rrx; ++i, ++i1)
                   for (j=trans_VQ->rry-1, j1=0; j>=(int)trans_VQ->ry; --j, ++j1) 
                       COMPUTE_LOOP_VQ
                   break;
           case 6: for (j=trans_VQ->rry-1,i1 = 0; j>=(int)trans_VQ->ry; --j, ++i1)
                   for (i=trans_VQ->rrx-1,j1=0; i>=(int)trans_VQ->rx; --i, ++j1) 
                       COMPUTE_LOOP_VQ
                   break;
           case 7: for (i=trans_VQ->rrx-1,i1 = 0; i>=(int)trans_VQ->rx; --i, ++i1)
                   for (j=trans_VQ->ry, j1=0; j<trans_VQ->rry; ++j, ++j1) 
                       COMPUTE_LOOP_VQ
                   break;
        }  
    }

    printf("\n%d transformations (VQ) applied.",count);
}


/**************************************************************/
/* APPLY THE TRANSFORMATIONS once for the FRACTAL part        */
/**************************************************************/
apply_transformations()
{
    IMAGE_TYPE **tempimage;
    int	i,j,i1,j1,count=0,domsize,w2;
    double pixel, mean;
    
    trans = &transformations;
    
    while (trans->next != NULL) {
        trans = trans->next;
	++count;
	domsize = (trans->rrx-trans->rx)*2;
	w2 = domsize * domsize;
	mean = 0;
	
	for(i=trans->dx; i < trans->dx + domsize; i++)
	  for(j=trans->dy; j < trans->dy + domsize; j++)
	    mean+=image[j][i];
	mean = mean/w2;

/* Since the inner loop is the same in each case of the switch below */
/* we just define it once for easy modification.                     */
#define COMPUTE_LOOP              {                                  \
        pixel = (image[j1][i1]+image[j1][i1+1]+image[j1+1][i1]+      \
                image[j1+1][i1+1])/4.0-mean;                         \
        image1[j][i] = bound(0.5 + trans->scale*pixel+trans->offset);\
        }

        switch(trans->sym_op) {
           case 0: for (i=trans->rx, i1 = trans->dx; 
                         i<trans->rrx; ++i, i1 += 2)
                   for (j=trans->ry, j1 = trans->dy; 
                         j<trans->rry; ++j, j1 += 2) 
                       COMPUTE_LOOP
                   break;
           case 1: for (j=trans->ry, i1 = trans->dx;
                           j<trans->rry; ++j, i1 += 2)
                   for (i=trans->rrx-1,
                           j1 = trans->dy; i>=(int)trans->rx; --i, j1 += 2) 
                       COMPUTE_LOOP
                   break;
           case 2: for (i=trans->rrx-1, 
                          i1 = trans->dx; i>=(int)trans->rx; --i, i1 += 2)
                   for (j=trans->rry-1,
                           j1 = trans->dy; j>=(int)trans->ry; --j, j1 += 2) 
                       COMPUTE_LOOP
                   break;
           case 3: for (j=trans->rry-1,
                           i1 = trans->dx; j>=(int)trans->ry; --j, i1 += 2)
                   for (i=trans->rx, j1 = trans->dy;
                           i<trans->rrx; ++i, j1 += 2) 
                       COMPUTE_LOOP
                   break;
           case 4: for (j=trans->ry, i1 = trans->dx;
                           j<trans->rry; ++j, i1 += 2)
                   for (i=trans->rx, j1 = trans->dy;
                           i<trans->rrx; ++i, j1 += 2) 
                       COMPUTE_LOOP
                   break;
           case 5: for (i=trans->rx, i1 = trans->dx;
                           i<trans->rrx; ++i, i1 += 2)
                   for (j=trans->rry-1,
                           j1 = trans->dy; j>=(int)trans->ry; --j, j1 += 2) 
                       COMPUTE_LOOP
                   break;
           case 6: for (j=trans->rry-1,
                           i1 = trans->dx; j>=(int)trans->ry; --j, i1 += 2)
                   for (i=trans->rrx-1,
                           j1 = trans->dy; i>=(int)trans->rx; --i, j1 += 2) 
                       COMPUTE_LOOP
                   break;
           case 7: for (i=trans->rrx-1, 
                          i1 = trans->dx; i>=(int)trans->rx; --i, i1 += 2)
                   for (j=trans->ry, j1 = trans->dy; 
                          j<trans->rry; ++j, j1 += 2) 
                       COMPUTE_LOOP
                   break;
        }  
  
    }
    tempimage = image;
    image = image1;
    image1 = tempimage;

    printf("\n%d transformations (fractal) applied.",count);
}

/*****************************************************************/
/* BUILD_HUFFMAN_TREE recursively generates a binary huffmantree */ 
/* by reading all single codewords. When a codeword is reached   */
/* at its end we can mark the corresponding huffmantree node as  */
/* leaf and save the codeword index (for VQ scal. factor).       */
/*****************************************************************/
build_huffman_tree(struct huffman_tree_node *node, char *codeword,
		   int n, int depth)
{
  int i;
  struct huffman_tree_node *left_son, *right_son;
  
  if (node->type==LEAF && codeword[depth]!='\0')  /* we must ADD TWO NEW LEAVES */
    {
      left_son  = (struct huffman_tree_node*)
	          malloc(sizeof(struct huffman_tree_node));
      right_son = (struct huffman_tree_node*)
	          malloc(sizeof(struct huffman_tree_node));

      /* UPDATE TYPES AND POINTERS of temporary node and its sons 
	 in binary tree */
      left_son->type   = LEAF;
      right_son->type  = LEAF;
      left_son->left   = NULL;
      left_son->right  = NULL;
      right_son->left  = NULL;
      right_son->right = NULL;
      node->type       = NODE;
      node->left       = left_son;
      node->right      = right_son;
	  
      if (codeword[depth]=='0') 
	build_huffman_tree(left_son, codeword, n, depth+1);
      else 
	build_huffman_tree(right_son, codeword, n, depth+1);
    }    
  else if (codeword[depth]=='\0') /* we reached end of code so 
				     STORE CODEBOOK INDEX */
    node->index = n;
  else
    { /* we pass an existing node, so PASS DEEPER instantly */ 
      if (codeword[depth]=='0') 
	build_huffman_tree(node->left, codeword, n, depth+1);
      else 
	build_huffman_tree(node->right, codeword, n, depth+1);
    } 
  return;
}

/*****************************************************************/
/* FRRE_HUFFMAN_TREE recursively frees allocated node memory.    */
/*****************************************************************/
free_huffman_tree(struct huffman_tree_node *node)
{
  if (node == NULL) return; 
  free_huffman_tree(node->left);
  free_huffman_tree(node->right);
  free(node);
}

/******************************************************************/
/* READ_SBITS reads in the variable length code used for scaling- */
/* factor information in OMSGVQ. We pass the huffman tree layer  */
/* by layer with direction given by unpack(next bit). We know     */
/* when a final leaf is reached out of type-information in        */
/* <huff_tree> data structure. Then unpacking can be stoped and   */
/* the index information can be returned.                         */
/******************************************************************/
read_sbits(struct huffman_tree_node *node, int *index)
{
  int bit;
  
  if (node->type==NODE)
    {
      bit=(int)unpack(1,2,input);
      if (bit==0)
	read_sbits(node->left, index);
      else 
	read_sbits(node->right, index);
    }
  else 
    *index=node->index;

  return;
}

/*************************************************************/
/*        This should really be done when they are read in.  */
/*************************************************************/
scale_transformations(double scalefactor)
{
    trans = &transformations;
    while (trans->next != NULL) {
        trans = trans->next;

		trans->rrx *= scalefactor;
		trans->rry *= scalefactor;
		trans->rx *= scalefactor;
		trans->ry *= scalefactor;
		trans->dx *= scalefactor;
		trans->dy *= scalefactor;
    }
}

/*****************************************************************/
/* RECURSIVELY PARTITION AN IMAGE, finding the largest contained */
/* square and call read_maindata .                               */
/*****************************************************************/
partition_image(atx, aty, hsize,vsize )
int atx, aty, hsize,vsize;
{
   int x_exponent,    /* the largest power of 2 image size that fits */
       y_exponent,    /* horizontally or vertically the rectangle.   */
       exponent,      /* The actual size of image that's encoded.    */
       size, 
       depth;
   
   x_exponent = (int)floor(log((double)hsize)/log(2.0));
   y_exponent = (int)floor(log((double)vsize)/log(2.0));
   
   /* exponent is min of x_ and y_ exponent */
   exponent = (x_exponent > y_exponent ? y_exponent : x_exponent);
   size = 1<<exponent; 
   depth = max_exponent - exponent;

   /* SET COUNTER DATA to zero */
   counter.mqueue=0; counter.oqueue=0;

   /* START UNPACKING all the rest out of inputfile */
   /* unpack(-2,2,input); */
   read_maindata(atx,aty,size,size,depth);

   if (size != hsize) 
      partition_image(atx+size, aty, hsize-size,vsize );

   if (size != vsize) 
      partition_image(atx, aty+size, size,vsize-size );
}        
  
/***************************************************************/
/* Scan the image and AVERAGE the transformation BOUNDARIES.   */
/***************************************************************/
smooth_image()
{
    IMAGE_TYPE pixel1, pixel2;
	int i,j;
	int w1,w2;

	printf("\nPostprocessing Image.");
    trans = &transformations;
    while (trans->next != NULL) {
        trans = trans->next;
		if (trans->rx == 0 || trans->ry == 0) 
            continue;
		
		if (trans->depth == max_part) {
			w1 = 5;
			w2 = 1;
		} else {
			w1 = 2;
			w2 = 1;
		}

        for (i=trans->rx; i<trans->rrx; ++i) {
             pixel1 = image[(int)trans->ry][i];
             pixel2 = image[(int)trans->ry-1][i];
             image[(int)trans->ry][i] = (w1*pixel1 + w2*pixel2)/(w1+w2);
             image[(int)trans->ry-1][i] = (w2*pixel1 + w1*pixel2)/(w1+w2);
        }

        for (j=trans->ry; j<trans->rry; ++j) {
             pixel1 = image[j][(int)trans->rx];
             pixel2 = image[j][(int)trans->rx-1];
             image[j][(int)trans->rx] = (w1*pixel1 + w2*pixel2)/(w1+w2);
             image[j][(int)trans->rx-1] = (w2*pixel1 + w1*pixel2)/(w1+w2);
        }
    }

    trans_VQ = &transformations_VQ;
    while (trans_VQ->next != NULL) {
        trans_VQ = trans_VQ->next;
		if (trans_VQ->rx == 0 || trans_VQ->ry == 0) 
            continue;
		
		if (trans_VQ->depth == max_part) {
			w1 = 5;
			w2 = 1;
		} else {
			w1 = 2;
			w2 = 1;
		}

        for (i=trans_VQ->rx; i<trans_VQ->rrx; ++i) {
             pixel1 = image[(int)trans_VQ->ry][i];
             pixel2 = image[(int)trans_VQ->ry-1][i];
             image[(int)trans_VQ->ry][i] = (w1*pixel1 + w2*pixel2)/(w1+w2);
             image[(int)trans_VQ->ry-1][i] = (w2*pixel1 + w1*pixel2)/(w1+w2);
        }

        for (j=trans_VQ->ry; j<trans_VQ->rry; ++j) {
             pixel1 = image[j][(int)trans_VQ->rx];
             pixel2 = image[j][(int)trans_VQ->rx-1];
             image[j][(int)trans_VQ->rx] = (w1*pixel1 + w2*pixel2)/(w1+w2);
             image[j][(int)trans_VQ->rx-1] = (w2*pixel1 + w1*pixel2)/(w1+w2);
        }
    }
}






