/******************************************************************************
 *
 * NAME
 *    encode.c
 * 
 * VERSION
 *    1.0
 *   
 * DATE
 *    10.9.1998
 * 
 *
 * IMPLEMENTATION 
 *    Bertram Ganz (ganz@informatik.uni-freiburg.de)
 *    Martin Mueller (mueller@www.fr.bw.schule.de)
 *    
 *    The fractal part is based on Yuval Fishers enc.c-code (yfisher@ucsd.edu)
 *
 *
 * SYNOPSIS    
 *    encode -[options] [inputfile [outputfile]]
 *
 * DESCRIPTION
 *    encode a byte image using an optimal hybrid scheme (fractal, oriented mean
 *    removed shape gain VQ) with a quadtree partition
 *   
 * OPTIONS
 *    (# = number), defaults show in ()
 *    -l # lambda multiplier for fidelity. (0.0)
 *    -m # minimum quadtree partitions. (4)
 *    -M # maximum quadtree partitions. (6)
 *    -S # number of input bytes to ignore. (2147429736)
 *    -w # width (horizontal size) of input data. (512)
 *    -h # height (vertical size) of input data. (512)
 *    -x # horizontal number of cluster centers. (12)
 *    -y # vertical number of cluster centers. (6)
 *    -Z # number of centers beeing visited for nn-search (1)
 *    -z # best of positive/negative phi's (on)(1)
 *    -d # domain step size. (1)
 *    -D # method {0,1,2} for domain pool (0)
 *    -s # number of scaling quantizing bits. (5)
 *    -o # number of offset quantizing bits. (7)
 *    -N # maximum scaling in encoding. (1.0)
 *    -V4 # maximum scaling in VQ encoding (4x4). (250.0)
 *    -V8 # maximum scaling in VQ encoding (8x8). (500.0)
 *    -V9 # maximum scaling in VQ encoding (16x16). (1000.0)
 *    -e   domain step as multiplier not divisor. (off)
 *    -p   use only positive scaling (for speed). (off)
 *
 *           See enc.h for definitions of default values
 *
 ******************************************************************************/

#include "encode.h"

int main(int argc, char **argv)
{
    /*** Defaults are set initially ***/
    double            lambda;         /* Lagrange multiplier  */
    char              inputfilename[200],
                      outputfilename[200];
    int               i,
                      hsize = -1,    /* The horizontal and vertical    */
                      vsize = -1;    /* size of the input image.       */     
    long              stripchar=0;   /* chars to ignore in input file. */
    FILE              *input; 

    
    /* START timer */
    START;       

    inputfilename[0]  = 1; /* We initially set the input to this and */
    outputfilename[0] = 1; /* then check if the input/output names   */
                           /* have been set below.                   */
    
    /* INITIALIZATION routine of global variables */
    initialize();          

    /**** SCAN through the input line and read in the arguments ****/
    read_commandline(&argc, argv, &lambda, inputfilename, outputfilename,
		     &hsize, &vsize, &stripchar);
    
    if (inputfilename[0] == 1) strcpy(inputfilename, "lenna.raw");
    if (outputfilename[0] == 1) strcpy(outputfilename, "lenna.trn");
   
    if (hsize == -1)
        if (vsize == -1) hsize = vsize = 256;
        else hsize = vsize;
    else 
        if (vsize == -1) vsize = hsize;

    /* ALLOCATE MEMORY for the input image. Allocating one chunck saves  */
    /* work and time later.                                              */
    matrix_allocate(image, hsize, vsize, IMAGE_TYPE);
    matrix_allocate(domimage[0], hsize/2, vsize/2, double);
    matrix_allocate(domimage[1], hsize/2, vsize/2, double);
    matrix_allocate(domimage[2], hsize/2, vsize/2, double);
    matrix_allocate(domimage[3], hsize/2, vsize/2, double);

    /* max_ & min_ part are variable, so this must be run time allocated */
    bits_needed = (int *)malloc(sizeof(int)*(1+max_part-min_part));

    if ((input = fopen(inputfilename, "r")) == NULL)
        fatal("Can't open input file.");

    /* SKIP the first  stripchar  chars */ 
    fseek(input, stripchar, 0); 
    i = fread(image[0], sizeof(IMAGE_TYPE), hsize*vsize, input);
    fclose(input);
    
    /* OUTPUT COMMENT on input status */
    if (i < hsize*vsize) 
        fatal("Not enough input data in the input file.");
    else
        printf("\n%dx%d=%d pixels read from %s.\n", hsize,vsize,i,inputfilename);
    fflush(stdout);

    /* MEMORY ALLOCATION for the list of clusters */
    size_clus = c_x * c_y;
    dim_clus = (int *)malloc(sizeof(int)*(1+max_part-min_part));
    nn       = (int *)malloc(sizeof(int)*(1+max_part-min_part));
    for (i=0; i<=max_part-min_part; ++i)
      dim_clus[i] = size_clus-1;
    
    list_perm = (int **)malloc((max_part-min_part+1) * sizeof(int *));
    for (i=0; i<=max_part-min_part; ++i)
      list_perm[i] = (int *)malloc(size_clus * sizeof(int));

    if(!best)
      {
	list_dist = (double *)malloc(2*number_nn_class * sizeof(double));
	list_clu  = (int *)malloc(2*number_nn_class * sizeof(int));
	list_sgn   = (int *)malloc(2*number_nn_class * sizeof(int));
      }
    else
      {
	list_dist = (double *)malloc(number_nn_class * sizeof(double));
	list_clu  = (int *)malloc(number_nn_class * sizeof(int));
	list_sgn   = (int *)malloc(number_nn_class * sizeof(int));
      }
 
    /**** ALLOCATE memory for DOMAIN DATA and INITIALIZE it ****/
    compute_sums(hsize, vsize);

    /* ENCODE THE IMAGE WITH HYBRID SCHEME */
    encode_image(hsize, vsize, lambda);
    
    /* OUTPUT DATA in outputfile */
    merge_outputdata(outputfilename, hsize, vsize);

    /*** FREE allocated memory ***/
    free(bits_needed);
    free(domimage);
    free(domimage[1]);
    free(domimage[2]);
    free(domimage[3]);
    free(domain.no_h_domains);
    free(domain.no_v_domains);
    free(domain.domain_hsize);
    free(domain.domain_vsize);
    free(domain.domain_hstep);
    free(domain.domain_vstep);
    for (i=0; i <= max_part-min_part; ++i) 
      free(domain.pixel[i]);   
    free(domain.pixel);
    free(image[0]);
    free(centers);
    free(code);
   
    /* STOP timer */
    STOP;       
    seconds=tmsend.tms_utime-tmsstart.tms_utime;
    
    printf("CPU time: %7.2f sec\n",
	   seconds/(double)CLK_TCK);
    
    return(0);
}

/******************************************************************/
/* Take a region of the image at x,y and CLASSIFY it.             */
/* The four quadrants of the region are ordered from brightest to */
/* least bright average value, then it is rotated into one of the */
/* three canonical orientations possible with the brightest quad  */
/* in the upper left corner.                                      */
/* The routine returns indices wich are: class number (pfirst);   */
/* the symmetry operation that bring the square into canocical    */
/* position; and the sum and sum^2 of the pixel values            */
/******************************************************************/
void classify(int x, int y, int xsize, int ysize, int *pfirst, double **pphi,
	      int *psym, double *psum, double *psum2, int type)
{
    int order[4], i,j,k,l=0, m, n; 
    double a[4],a2[4], *vec, *temp;
    void (*average_func)();
    
    if (type == 2) average_func = average; else average_func = average1;

    /* GET THE AVERAGE VALUES of each quadrant */
    (*average_func)(x,y,                 xsize/2,ysize/2,   &a[0], &a2[0]);
    (*average_func)(x,y+ysize/2,         xsize/2,ysize/2,   &a[1], &a2[1]);
    (*average_func)(x+xsize/2,y+ysize/2, xsize/2,ysize/2,   &a[2], &a2[2]);
    (*average_func)(x+xsize/2,y,         xsize/2,ysize/2,   &a[3], &a2[3]);
    *psum = a[0] + a[1] + a[2] + a[3];
    *psum2 =  a2[0] + a2[1] + a2[2] + a2[3];

    for (i=0; i<4; ++i) 
      {
	/* after the sorting below order[i] is the ith brightest */
	/* quadrant.                                              */
	order[i] = i; 
	/* CONVERT a2[] to store the variance of each quadrant    */
	a2[i] -= (double)(1<<(2*type))*a[i]*a[i]/(double)(xsize*ysize);
      }
    
    /* Now ORDER THE AVERAGE VALUE and also in order[],   which will */
    /* then tell us the indices (in a[]) of the brightest to darkest */
    for (i=2; i>=0; --i)
    for (j=0; j<=i; ++j)
      if (a[j]<a[j+1]) {
	  swap(order[j], order[j+1],int)
          swap(a[j], a[j+1],double)
    }
    
    /* because of the way we ordered the a[] the rotation can be */
    /* read right off of order[]. That will make the brightest   */
    /* quadrant be in the upper left corner. But we must still   */ 
    /* decide which cannonical class the image portion belogs    */
    /* to and whether to do a flip or just a rotation. This is   */
    /* the following table summarizes the horrid lines below     */
    /* order      class            do a rotation                 */
    /* 0,2,1,3      0                   0                        */
    /* 0,2,3,1      0                   1                        */
    /* 0,1,2,3      1                   0                        */
    /* 0,3,2,1      1                   1                        */
    /* 0,1,3,2      2                   0                        */
    /* 0,3,1,2      2                   1                        */

    *psym = order[0]; 
    
    /* ROTATE the values */
    for (i=0; i<4; ++i)
        order[i] = (order[i] - (*psym) + 4)%4; 

    for (i=0; order[i] != 2; ++i); 
    *pfirst = i-1;
    if (order[3] == 1 || (*pfirst == 2 && order[2] == 1)) *psym += 4;

    vec=(double *)malloc(dim * sizeof(double));
 
    if(type == 2)  /* Domains */
      {
	k = ((x%2)<<1) + y%2;
	x >>= 1; y >>= 1;
	xsize >>= 1; ysize >>= 1;

	/* COMPUTE FEATURE VECTORS phi */
	temp=(double *)malloc(dim * sizeof(double));
	
	for (i=x; i<x+xsize; i+=(xsize*0.25))
	  for (j=y; j<y+ysize; j+=(ysize*0.25)) 
	    {
	      temp[l] = 0;
	      for(m=0; m<xsize*0.25; m++)    /* averaging */
		for(n=0; n<ysize*0.25; n++)
		  temp[l] += domimage[k][j+n][i+m];
	      temp[l] = temp[l]*16/(xsize*ysize);
	      l++;
	    }
	for(j=0;j<dim;j++)   /* rotate in canonical position */ 
	  vec[j] = temp[phi_trafo[*psym][j]];
	
	compute_phi(vec, pphi, dim, *psum*16/(xsize*ysize));
	free(temp);
      }

    else /* Ranges */
      {
	/* compute feature vectors phi */
	for (i=x; i<x+xsize; i+=(xsize*0.25))
	  for (j=y; j<y+ysize; j+=(ysize*0.25)) 
	    {
	      vec[l] = 0;
	      for(m=0; m<xsize*0.25; m++)    /* averaging */
		for(n=0; n<ysize*0.25; n++)
		  vec[l] += image[j+n][i+m];
	      vec[l] = vec[l]*16/(xsize*ysize);
	      l++;
	    }
	compute_phi(vec, pphi, dim, *psum*16/(xsize*ysize));
	/*phi not in canonical posititon! */
      }

    free(vec);
}


/*****************************************************************/
/* COMPUTE_SUMS reads in image data and computes domain vec-     */
/* tors. The SOM_PAK programm is executed for clustering.        */
/* After this we can assign every domain to its nearest cluster  */
/* center.                                                       */
/* Compute sum and sum^2 of pixel values in domains for use in   */
/* the se computation later. Since a domain is compared with     */
/* many ranges, doing this just once saves a lot of computation  */
/* This routine also fills a list structure with the domains     */
/* as they are classified and creates the memory for the domain  */
/* data in a matrix.                                             */
/*****************************************************************/
void compute_sums(int hsize, int vsize)
{
    int i,j,k,l,m,n,o,
        domain_x, 
        domain_y, 
        clu_nr,
        size,  
        x_exponent, 
        y_exponent,
	first_class;
    double qerror,
        vecnorm,
        *tmp;
    char ch[200], str[100];
    FILE   *centerlist;
    struct classified_domain *node;

    printf("\nComputing domain vectors ... ");
    fflush(stdout); 

    /* PREDECIMATE THE IMAGE into domimage to avoid having to      */
    /* do repeated averaging of 2x2 pixel groups.                   */
    /* There are 4 ways to decimate the image, depending on the     */
    /* location of the domain, odd or even address.                 */
    for (i=0; i<2; ++i)
      for (j=0; j<2; ++j)
	for (k=i; k<hsize-i; k += 2)
	  for (l=j; l<vsize-j; l += 2) 
	    domimage[(i<<1)+j][l>>1][k>>1] = 
	      ((double)image[l][k] + (double)image[l+1][k+1] +
	      (double)image[l][k+1] + (double)image[l+1][k])*0.25;

    /* ALLOCATE MEMORY for the sum and sum^2 of domain pixels        */
    /* We first COMPUTE THE SIZE OF THE LARGEST SQUARE that fits in  */
    /* the image.                                                    */
    x_exponent = (int)floor(log((double)hsize)/log(2.0));
    y_exponent = (int)floor(log((double)vsize)/log(2.0));
   
    /* exponent is min of x_ and y_ exponent */
    max_exponent = (x_exponent > y_exponent ? y_exponent : x_exponent);

    /* size is the size of the largest square that fits in the image */
    /* It is used to compute the domain and range sizes.             */
    size =  1<<max_exponent; 

    if (max_exponent < max_part)
      fatal("Reduce maximum number of quadtree partitions.");
    if (max_exponent-2 < max_part)
      printf("\nWarning: so many quadtree partitions yield absurd ranges.");

    i = max_part - min_part + 1;
    domain.no_h_domains = (int *)malloc(sizeof(int)*i);
    domain.no_v_domains = (int *)malloc(sizeof(int)*i);
    domain.domain_hsize = (int *)malloc(sizeof(int)*i);
    domain.domain_vsize = (int *)malloc(sizeof(int)*i);
    domain.domain_hstep = (int *)malloc(sizeof(int)*i);
    domain.domain_vstep = (int *)malloc(sizeof(int)*i);

    domain.pixel= (struct domain_pixels ***)
              malloc(i*sizeof(struct domain_pixels **));
    if (domain.pixel == NULL) fatal("No memory for domain pixel sums.");

    for (i=0; i <= max_part-min_part; ++i) 
      {
	/* first COMPUTE HOW MANY DOMAINS there are HORIZONTALLY */
	domain.domain_hsize[i] = size >> (min_part+i-1);
	if (dom_type == 2) 
	  domain.domain_hstep[i] = dom_step; 
	else if (dom_type == 1)
	  if (dom_step_type == 1) 
	    domain.domain_hstep[i] = (size >> (max_part - i-1))*dom_step;
	  else 
	    domain.domain_hstep[i] = (size >> (max_part - i-1))/dom_step;
	else 
	  if (dom_step_type == 1) 
	    domain.domain_hstep[i] = domain.domain_hsize[i]*dom_step;
	  else 
	    domain.domain_hstep[i] = domain.domain_hsize[i]/dom_step;
	
	domain.no_h_domains[i] = 1+(hsize-domain.domain_hsize[i])/
	  domain.domain_hstep[i];
	
	/* now COMPUTE HOW MANY DOMAINS there are vertically. The sizes  */
	/* are the same for square domains, but not for rectangular ones */
	domain.domain_vsize[i] = size >> (min_part+i-1);
	if (dom_type == 2) 
	  domain.domain_vstep[i] = dom_step; 
	else if (dom_type == 1)
	  if (dom_step_type == 1) 
	    domain.domain_vstep[i] = (size >> (max_part - i-1))*dom_step;
	  else
	    domain.domain_vstep[i] = (size >> (max_part - i-1))/dom_step;
	else 
	  if (dom_step_type == 1) 
	    domain.domain_vstep[i] = domain.domain_vsize[i]*dom_step;
	  else
	    domain.domain_vstep[i] = domain.domain_vsize[i]/dom_step;
	
	domain.no_v_domains[i] = 1+(vsize-domain.domain_vsize[i])/
	  domain.domain_vstep[i];
	
	/* Now COMPUTE the NUMBER OF BITS needed to store the domain data */
	bits_needed[i] = ceil(log((double)domain.no_h_domains[i]*
				  (double)domain.no_v_domains[i])/log(2.0)); 
	
	matrix_allocate(domain.pixel[i], domain.no_h_domains[i],
			domain.no_v_domains[i], struct domain_pixels);
	
	printf("\n%i %ix%i-Domains",domain.no_h_domains[i]*domain.no_v_domains[i],
	       domain.domain_vsize[i],domain.domain_vsize[i]);
	fflush(stdout); 
	
	/* Scaling factors for VQ */
	if ((max_part-min_part == 0) && (domain.domain_hsize[i] == 8))
	  max_scale_vq[1] = max_scale_vq[0];
	if ((max_part-min_part == 0) && (domain.domain_hsize[i] == 16))
	  max_scale_vq[0] = max_scale_vq[1];      
	if ((max_part-min_part == 0) && (domain.domain_hsize[i] == 32))
	  max_scale_vq[0] = max_scale_vq[2];      
      }
    
    if ((max_part-min_part != 0)&&(domain.domain_hsize[max_part-min_part] == 16))
      max_scale_vq[0] = max_scale_vq[1];      

    /* PARTITIONING of the domains        */
    /* COMPUTATION of FEATURE VECTORS phi */
    for (i=0; i <= max_part-min_part; ++i) 
      {
	for (j=0,domain_x=0; j<domain.no_h_domains[i]; ++j,
	       domain_x+=domain.domain_hstep[i]) 
	  for (k=0,domain_y=0; k<domain.no_v_domains[i]; ++k,
		 domain_y+=domain.domain_vstep[i])
	    {
	      domain.pixel[i][k][j].phi_vec = (double *)
		malloc(dim * sizeof(double));
	      if (domain.pixel[i][k][j].phi_vec == NULL) 
		fatal("\n not enough memory for domain feature vectors");
	      
	      classify(domain_x, domain_y, 
		       domain.domain_hsize[i], 
		       domain.domain_vsize[i], 
		       &first_class,   /* not used for domains */
		       &domain.pixel[i][k][j].phi_vec,
		       &domain.pixel[i][k][j].sym, 
		       &domain.pixel[i][k][j].sum, 
		       &domain.pixel[i][k][j].sum2, 2);
	      /* When the domain data is referenced from the list, we need to */
	      /* know where the domain is.. so we have to store the position  */
	      domain.pixel[i][k][j].dom_x = j;
	      domain.pixel[i][k][j].dom_y = k;
	    }
      }
    
    i = max_part - min_part + 1;
    
    centers  = (double ***) malloc(i * sizeof(double**)); 
    
    /* ALLOCATE and zero the list containing the classified domain data */
    /* Classification with cluster centers                              */
    the_domain = (struct classified_domain ***) 
	    malloc(size_clus * sizeof(struct classified_domain **));
    for (clu_nr = 0; clu_nr < (size_clus); ++clu_nr)
      {
	the_domain[clu_nr] = 
	  (struct classified_domain **) 
	    malloc(i*sizeof(struct classified_domain *));
	for (j=0; j<i; ++j)
	  the_domain[clu_nr][j] = NULL;
      }
    
    printf("\n\nClustering .....\n");
    fflush(stdout); 
    for (i=0; i <= max_part-min_part; ++i) 
      {      
	qerror = 0.0;
	
	if(domain.domain_hsize[i] == 16)
	  {
	    sprintf(str, "cen_64.som");
	    if ((centerlist = fopen(str, "r")) == NULL)
	      fatal("Can't open center file.");

	    /* INTIALIZATION of cluster list */
	    fgets(ch, 200, centerlist);  /* SKIP first line */
	    centers[i]  = (double **) malloc(size_clus * sizeof(double*)); 
	    centers_64  = (double **) malloc(size_clus * sizeof(double*)); 

	    /* READ IN cluster center vectors */
	    for(j=0; j<size_clus; j++)       
	      {
		centers[i][j] = (double *) malloc(dim * sizeof(double)); 
		centers_64[j] = (double *) malloc(64 * sizeof(double)); 

		for(k=0; k<64; k++)
		  if (fscanf(centerlist, "%lf", &centers_64[j][k])==EOF) 
		    goto ende;
		  /* break loop at end of file */    
		
		/* NORMALIZE  cluster center vectors */
		vecnorm = sqrt(norm(centers_64[j], 64));
		for(k=0; k<64; k++)
		  centers_64[j][k] = centers_64[j][k]/vecnorm;
		
		/* AVERAGING for 64-dim vector */
		for (k=0, m=0; k<64; k+=16)
		  for (l=0; l<8; l+=2, m++) 
		    {centers[i][j][m] = 
		      (centers_64[j][k+l] + centers_64[j][k+l+1] +
		       centers_64[j][k+l+8] + centers_64[j][k+l+9]) * 0.25;}
		    
		/* NORMALIZATION of 16-dim cluster center vectors */
		vecnorm = sqrt(norm(centers[i][j], dim));
		for(k=0; k<dim; k++)
		  centers[i][j][k] = centers[i][j][k]/vecnorm;

	      }
	  ende:
	    fclose(centerlist);

	    /* COMPUTE NUMBER OF BITS we use to store the index of 
	       CODEVECTOR */
	    vec_bits  = ceil(log(size_clus)/log(2.0)); 
	  }

	else if(domain.domain_hsize[i] == 32)
	  {
	    sprintf(str, "cen_256.som");
	    if ((centerlist = fopen(str, "r")) == NULL)
	      fatal("Can't open center file.");

	     /* INTIALIZATION of cluster list */
	    fgets(ch, 200, centerlist);         /* skip first line */
	    centers[i]  = (double **) malloc(size_clus * sizeof(double*)); 
	    centers_256  = (double **) malloc(size_clus * sizeof(double*)); 
	    
	    /* READ IN cluster center vectors */
	    for(j=0; j<size_clus; j++)       
	      {
		centers[i][j] = (double *) malloc(dim * sizeof(double)); 
		centers_256[j] = (double *) malloc(256 * sizeof(double)); 

		for(k=0; k<256; k++)
		  if (fscanf(centerlist, "%lf", &centers_256[j][k])==EOF) 
		    goto ende3;
		  /* break loop at end of file */    
		
		/* NORMALIZE cluster center vectors */
		vecnorm = sqrt(norm(centers_256[j], 256));
		for(k=0; k<256; k++)
		  centers_256[j][k] = centers_256[j][k]/vecnorm;

		/* AVERAGING for 256-dim clusters */
		for (k=0, m=0; k<256; k+=64)
		  for (l=0; l<16; l+=4, m++) 
		    {
		      centers[i][j][m] = 0;
		      for (n=0; n<4; n++) 
			for (o=0; o<64; o+=16) 
			  centers[i][j][m] += centers_256[j][k+l+n+o]; 
	
		      centers[i][j][m] *= 0.0625;
		    }	

		/* NORMALIZATION of 16-dim cluster centers */
		vecnorm = sqrt(norm(centers[i][j], dim));
		for(k=0; k<dim; k++)
		  centers[i][j][k] = centers[i][j][k]/vecnorm;
	      }
	  ende3:
	    
	    fclose(centerlist);

	    /* COMPUTE NUMBER OF BITS we use to store the 
	       index of codevector */
	    vec_bits  = ceil(log(size_clus)/log(2.0)); 
	  }

	else
	  {
	    sprintf(str, "cen_16.som");
	    if ((centerlist = fopen(str, "r")) == NULL)
	      fatal("Can't open center file.");

	    /* INTIALIZATION of cluster list */
	    fgets(ch, 200, centerlist);        /* SKIP first line */  
	    centers[i]  = (double **) malloc(size_clus * sizeof(double*)); 

	    /* READ IN cluster center vectors */
	    for(j=0; j<(size_clus); j++)      
	      {
		centers[i][j] = (double *) malloc(dim * sizeof(double)); 
		
		for(k=0; k<dim; k++)
		  if (fscanf(centerlist, "%lf", &centers[i][j][k])==EOF)
		    goto ende2;
		 /* break loop at end of file */    
		
		/* NORMALIZE cluster center vectors */
		vecnorm = sqrt(norm(centers[i][j], dim));
		for(k=0; k<dim; k++)
		  centers[i][j][k] = centers[i][j][k]/vecnorm;
	      }
	  ende2:
	    fclose(centerlist);

	    /* COMPUTE NUMBER OF BITS we use to store the 
	       index OF CODEVECTOR */
	    vec_bits  = ceil(log(size_clus)/log(2.0)); 
	  }
	
	/* ASSIGNMENT of domains to cluster centers */ 
	for (j=0; j<domain.no_h_domains[i]; ++j) 
	  for (k=0; k<domain.no_v_domains[i]; ++k)
	    if(domain.pixel[i][k][j].phi_vec)
	      {
		qerror += sqrt(nn_cluster_search(domain.pixel[i][k][j].phi_vec,
					      &clu_nr, 1, i));
		node = (struct classified_domain *) 
		  malloc(sizeof(struct classified_domain));
		node->the = &domain.pixel[i][k][j];
		node->next = the_domain[clu_nr][i];
		the_domain[clu_nr][i] = node;
	      }  
      }   

    /* Now we make sure no domain class is actually empty.  */
    /* REMOVE EMPTY CLUSTERS from list. */
    for (i=0; i <= max_part-min_part; ++i)
      {
	for(j=0; j<size_clus; j++)   /* INITIALIZATION of permutation list */
	  list_perm[i][j] = j;      /*-----------------------------------*/

	for(j=0; j < dim_clus[i]; j++)
	  {
	    while (the_domain[j][i] == NULL) 
	      {
		swap(list_perm[i][j], list_perm[i][dim_clus[i]], int);
		tmp=centers[i][dim_clus[i]];
		centers[i][dim_clus[i]]=centers[i][j];
		centers[i][j] = tmp;

		if (domain.domain_hsize[i] == 16) /* 64-dim  clusters have to */
		  {                               /* be permuted too.         */
		    tmp=centers_64[dim_clus[i]];
		    centers_64[dim_clus[i]]=centers_64[j];
		    centers_64[j] = tmp;
		  }

		if (domain.domain_hsize[i] == 32) /* 256-dim  clusters have to */
		  {                               /* be permuted too.          */
		    tmp=centers_256[dim_clus[i]];
		    centers_256[dim_clus[i]]=centers_256[j];
		    centers_256[j] = tmp;
		  }
	
		node=the_domain[dim_clus[i]][i];
		the_domain[dim_clus[i]][i]=the_domain[j][i];
		the_domain[j][i]=node;
		dim_clus[i]--;
	      }
	  }
	j=0;
	while ((j < size_clus) && (the_domain[j][i] != NULL) ) 
	  j++;		
	dim_clus[i]=j;

	if (dim_clus[i] < number_nn_class)
	  nn[i] = dim_clus[i];
	else
	  nn[i] = number_nn_class;
	if (!best)
	  nn[i] <<= 1;
      }
    
    printf("Done.\n\n");
    fflush(stdout); 
}

/******************************************************************/
/* COMPARE A RANGE TO A DOMAIN and return se and the quantized    */
/* scaling and offset values (pialhpa, pibeta).                   */
/* In order to compute beta and alpha independantly we orthogona- */
/* lize the domainblocks. In this case beta is the mean value of  */
/* the corresponding rangeblock.                                  */
/******************************************************************/
double compare_fr(int atx, int aty, int xsize, int ysize, int depth, 
	       double rsum, double rsum2, int dom_x, int dom_y, 
	       int sym_op, int *pialpha, int *pibeta)
{
    int i, j, i1, j1, k, 
        domain_x, 
        domain_y;        /* The domain position                */

    double pixel,
           det,          /* determinant of solution */
           dsum,         /* sum of domain values */
           rdsum = 0,    /* sum of range*domain values   */
           dsum2,        /* sum of domain^2 values   */
           w2 = 0,       /* total number of values tested */
           se = 0,       /* square difference */
           alpha,        /* the scale factor */
           beta;         /* The offset */

    w2 = xsize * ysize;

    dsum = domain.pixel[depth-min_part][dom_y][dom_x].sum;
    dsum2 = domain.pixel[depth-min_part][dom_y][dom_x].sum2;
    domain_x = (dom_x * domain.domain_hstep[depth-min_part]);
    domain_y = (dom_y * domain.domain_vstep[depth-min_part]);
    k = ((domain_x%2)<<1) + domain_y%2;
    domain_x >>= 1;
    domain_y >>= 1;

    /* The main statement in this routine is a switch statement which */
    /* scans through the domain and range to compare them. The loop   */
    /* center is the same so we #define it for easy modification      */
#define COMPUTE_LOOP              {                                 \
        pixel = domimage[k][j1][i1];                                \
        rdsum += image[j][i]*pixel;                                 \
        }
    
      switch(sym_op) {
      case 0: for (i=atx, i1 = domain_x; i<atx+xsize; ++i, ++i1)
	for (j=aty, j1 = domain_y; j<aty+ysize; ++j, ++j1) 
	  COMPUTE_LOOP
	    break;
      case 1: for (j=aty, i1 = domain_x; j<aty+ysize; ++j, ++i1)
	for (i=atx+xsize-1, j1 = domain_y; i>=atx; --i, ++j1) 
	  COMPUTE_LOOP
	    break;
      case 2: for (i=atx+xsize-1, i1 = domain_x; i>=atx; --i, ++i1)
	for (j=aty+ysize-1, j1 = domain_y; j>=aty; --j, ++j1) 
	  COMPUTE_LOOP
	    break;
      case 3: for (j=aty+ysize-1, i1 = domain_x; j>=aty; --j, ++i1)
	for (i=atx, j1 = domain_y; i<atx+xsize; ++i, ++j1) 
	  COMPUTE_LOOP
	    break;
      case 4: for (j=aty, i1 = domain_x; j<aty+ysize; ++j, ++i1)
	for (i=atx, j1 = domain_y; i<atx+xsize; ++i, ++j1) 
	  COMPUTE_LOOP
	    break;
      case 5: for (i=atx, i1 = domain_x; i<atx+xsize; ++i, ++i1)
	for (j=aty+ysize-1, j1 = domain_y; j>=aty; --j, ++j1) 
	  COMPUTE_LOOP
	    break;
      case 6: for (j=aty+ysize-1, i1 = domain_x; j>=aty; --j, ++i1)
	for (i=atx+xsize-1, j1 = domain_y; i>=atx; --i, ++j1) 
	  COMPUTE_LOOP
	    break;
      case 7: for (i=atx+xsize-1, i1 = domain_x; i>=atx; --i, ++i1)
	for (j=aty, j1 = domain_y; j<aty+ysize; ++j, ++j1) 
	  COMPUTE_LOOP
	    break;
      }
    
    det = (xsize*ysize)*dsum2 - dsum*dsum;
    
    if (det == 0.0) 
      alpha = 0.0; 
    else 
      alpha = (w2*rdsum - rsum*dsum)/det;
        
    if  (only_positive && alpha < 0.0) 
      alpha = 0.001;
    *pialpha = 0.5 + (alpha + max_scale)/(2.0*max_scale)*((1<<s_bits)-1);
    if (*pialpha < 0) *pialpha = 0;
    if (*pialpha >= 1<<s_bits) *pialpha = (1<<s_bits)-1;

    /* Now RECOMPUTE ALPHA back */
    alpha = (double)*pialpha/(double)((1<<s_bits)-1)*(2.0*max_scale)-max_scale;

    /* COMPUTE the OFFSET */
    beta = rsum/w2;
    
    /* CONVERT BETA to an integer */
    /* we use the sign information of alpha to pack efficiently */
    *pibeta = 0.5 + beta/(GREY_LEVELS)*((1<<o_bits)-1);
    
    /* RECOMPUTE BETA from the integer */
    beta = (double)*pibeta/(double)((1<<o_bits)-1)*
      (GREY_LEVELS);
    
    /* COMPUTE SQUARE ERROR based on the quantized alpha and beta! */
    se = rsum2 -2.0*alpha*rdsum+rsum*2.0*(alpha*dsum/w2-beta)
         +alpha*alpha*(dsum2-dsum*dsum/w2)+w2*beta*beta;
        
    return(se);
}        

/**************************************************************/
/* COMPARE A RANGE to a  CLUSTER CENTER                       */
/**************************************************************/
double compare_cen(int atx, int aty, int xsize, int ysize, int depth, 
		   double rsum, double rsum2,
		   int index,
     		   int sym_op, int *pialpha, int *pibeta, 
		   double *alpha, double *beta)
{
    int i, j, i1;
    double **cen_point,
           pixel,
           rdsum = 0,    /* sum of range*domain values   */
           w2 = 0,       /* total number of values tested */
           se = 0;       /* square difference */

    w2 = xsize * ysize;

    /* The main statement in this routine is a switch statement which */
    /* scans through the domain and range to compare them. The loop   */
    /* center is the same so we #define it for easy modification      */
#define COMPUTE_LOOP2              {                                 \
        pixel = centers[depth-min_part][index][i1];                 \
        rdsum += image[j][i]*pixel;                                 \
        }
    
    if(xsize == 4)
      {
	switch(sym_op) {
	case 0: for (i=atx, i1 = 0; i<atx+xsize; ++i)
	  for (j=aty; j<aty+ysize; ++j, ++i1) 
	    COMPUTE_LOOP2
	      break;
	case 1: for (j=aty, i1 = 0; j<aty+ysize; ++j)
	  for (i=atx+xsize-1; i>=atx; --i, ++i1) 
	    COMPUTE_LOOP2
	      break;
	case 2: for (i=atx+xsize-1, i1 = 0; i>=atx; --i)
	  for (j=aty+ysize-1; j>=aty; --j, ++i1) 
	    COMPUTE_LOOP2
	      break;
	case 3: for (j=aty+ysize-1, i1 = 0; j>=aty; --j)
	  for (i=atx; i<atx+xsize; ++i, ++i1) 
	    COMPUTE_LOOP2
	      break;
	case 4: for (j=aty, i1 = 0; j<aty+ysize; ++j)
	  for (i=atx; i<atx+xsize; ++i, ++i1) 
	    COMPUTE_LOOP2
	      break;
	case 5: for (i=atx, i1 = 0; i<atx+xsize; ++i)
	  for (j=aty+ysize-1; j>=aty; --j, ++i1) 
	    COMPUTE_LOOP2
	      break;
	case 6: for (j=aty+ysize-1, i1 = 0; j>=aty; --j)
	  for (i=atx+xsize-1; i>=atx; --i, ++i1) 
	    COMPUTE_LOOP2
	      break;
	case 7: for (i=atx+xsize-1, i1 = 0; i>=atx; --i)
	  for (j=aty; j<aty+ysize; ++j, ++i1) 
	    COMPUTE_LOOP2
	      break;
	}
      }
    else
#define COMPUTE_LOOP3              {                               \
        pixel = cen_point[index][i1];                               \
        rdsum += image[j][i]*pixel;                                  \
        }
    
      {
	if (xsize == 8)
	  cen_point = centers_64;
	else 
	  cen_point = centers_256;

	switch(sym_op) {
	case 0: for (i=atx, i1 = 0; i<atx+xsize; ++i)
	  for (j=aty; j<aty+ysize; ++j, ++i1) 
	    COMPUTE_LOOP3
	      break;
	case 1: for (j=aty, i1 = 0; j<aty+ysize; ++j)
	  for (i=atx+xsize-1; i>=atx; --i, ++i1) 
	    COMPUTE_LOOP3
	      break;
	case 2: for (i=atx+xsize-1, i1 = 0; i>=atx; --i)
	  for (j=aty+ysize-1; j>=aty; --j, ++i1) 
	    COMPUTE_LOOP3
	      break;
	case 3: for (j=aty+ysize-1, i1 = 0; j>=aty; --j)
	  for (i=atx; i<atx+xsize; ++i, ++i1) 
	    COMPUTE_LOOP3
	      break;
	case 4: for (j=aty, i1 = 0; j<aty+ysize; ++j)
	  for (i=atx; i<atx+xsize; ++i, ++i1) 
	    COMPUTE_LOOP3
	      break;
	case 5: for (i=atx, i1 = 0; i<atx+xsize; ++i)
	  for (j=aty+ysize-1; j>=aty; --j, ++i1) 
	    COMPUTE_LOOP3
	      break;
	case 6: for (j=aty+ysize-1, i1 = 0; j>=aty; --j)
	  for (i=atx+xsize-1; i>=atx; --i, ++i1) 
	    COMPUTE_LOOP3
	      break;
	case 7: for (i=atx+xsize-1, i1 = 0; i>=atx; --i)
	  for (j=aty; j<aty+ysize; ++j, ++i1) 
	    COMPUTE_LOOP3
	      break;
	}
      }
    

    *alpha = rdsum;
        
    if  (only_positive && *alpha < 0.0) *alpha = 0.0;
    *pialpha = 0.5 + (*alpha + max_scale_vq[max_part-depth])/
               (2.0* max_scale_vq[max_part-depth])*((1<<s_bits)-1);
    if (*pialpha < 0) *pialpha = 0;
    if (*pialpha >= 1<<s_bits) *pialpha = (1<<s_bits)-1;

    /* Now RECOMPUTE ALPHA back */
    *alpha = (double)*pialpha/(double)((1<<s_bits)-1)*
             (2.0*max_scale_vq[max_part-depth])-max_scale_vq[max_part-depth];
    
    /* COMPUTE the OFFSET */ 
     *beta = rsum/w2;
    
    /* CONVERT BETA to an integer */    
     *pibeta = 0.5 + *beta/(GREY_LEVELS)*((1<<o_bits)-1);
    
    /* RECOMPUTE BETA from the integer */
     *beta = (double)*pibeta/(double)((1<<o_bits)-1)*
      (GREY_LEVELS);
    
    /* COMPUTE SQUARE ERROR based on the quantized alpha and beta! */
    se=(rsum2 + *alpha*(*alpha - 2.0*rdsum) +
	    *beta*( *beta *w2 - 2.0*rsum)); 
   
    return(se);   
}

/****************************************************************/
/* ENCODE_IMAGE is the central procedure of our hybrid rate-    */
/* distortion optimized encoding scheme. First we optimally en- */
/* code all image blocks on every quadtree-level with the       */
/* hybrid scheme by minimizing lambda*rate +dist                */
/* All relevant data is saved in the quadtree structure         */
/* <quadtree_node>. Then a bottom-up traversal of the quadtree  */
/* is done for optimal pruning. A node is marked as final       */
/* leave when delta_dist/delta_rate<= lambda. After this we must */
/* store the temporary quadtree-data-structure in <coder_data>  */
/* for a recursively generated output file.                     */ 
/****************************************************************/
void encode_image(int hsize, int vsize, double lambda)
{
  int atx, aty;
  int max_blocksize;
  int i, depth;
  int tot_nob, temp_nob=0; /* total number of max_blocksize image blocks */
  double seconds,countdown;
  struct tms tmsstart1, tmsend1;  
  struct quadtree_node *root;

  /* ALLOCATE MEMORY for the complete code. Every quadtreelevel has its 
     own matrix. */
  depth=max_part-min_part+1;
  code = (struct coder_data ***) malloc(depth*sizeof(struct coder_data **));
  if (code == NULL) fatal("No memory for image code.");
  for (i=0; i <= depth; ++i) 
    matrix_allocate(code[i],1<<(min_part+i),1<<(min_part+i), struct coder_data);
      
  /* INITIALIZE and ALLOCATE quadtree for temporary data */
  root = (struct quadtree_node*) malloc(sizeof(struct quadtree_node));
  quadtree_alloc(root,depth);

  /* now START BLOCKWISE ENCODING of image at every quadtree level */
  max_blocksize = hsize/(1<<min_part); 
  tot_nob = hsize*vsize/(max_blocksize*max_blocksize);
  
  /* start timer */
  if (times(&tmsstart1)==-1) printf("times error");
    
  for(atx=0; atx<hsize; atx+=max_blocksize)
    for(aty=0; aty<vsize; aty+=max_blocksize)
      {
	/* the temporary quadtree is completely filled with optimal 
	   hybrid coder data */ 
	encode_quadtree(atx,aty,max_blocksize,min_part,lambda,root);
	
	/* PRUNE the whole QUADTREE using lambda criterion (bottom-up pruning) */
	for (i=1; i <= max_part-min_part; i++)
	  prune_quadtree(lambda,root,min_part,max_part-i);
	
	/* STORE DATA in data structure <code> */
	store_qtdata(atx,aty,root,min_part,max_blocksize);
	
	/* Check actual time and UPDATE TIME PREDICTION */
	if (times(&tmsend1)==-1) printf("times error");
	seconds=(tmsend1.tms_utime-tmsstart1.tms_utime)/(double)CLK_TCK;
	++temp_nob;
	countdown=seconds*((double)tot_nob/(double)temp_nob-1.);
	printf("\rEncoding Image .....%7.1f sec remaining",countdown);
	fflush(stdout);
      }
}

/***********************************************************************/
/* ENCODE_QUADTREE recursively encodes the temporary image segment     */
/* (atx,aty,max_blocksize) with an optimal hybrid scheme.              */
/* All relevant data used later for tree pruning is stored in root.    */
/***********************************************************************/
void encode_quadtree(int atx, int aty, int blocksize, int depth, double lambda,
		     struct quadtree_node *node)
{
  int i, j, m, n, l, 
      best_method,
      sym_op,                   /* A symmetry operation of the square */
      ialpha,                   /* Intgerized scalling factor         */
      ibeta,                    /* Intgerized offset                  */
      best_ialpha,              /* best ialpha found                  */
      best_ibeta,
      best_ibeta_c,             
      best_sym_op,
      best_ialpha_vq,       
      best_ibeta_vq,
      best_sym_op_vq,
      best_domain_x,
      best_domain_y, 
      best_vec,
      the_first_class,
      range_sym_op[2],          /* the operations to bring square to  */
      domain_sym_op;            /* cannonical position.               */
  double objective_val[3]       /* distortion (se) * lambda + rate     */
          ={HUGE,HUGE,HUGE};      
  double se,                    /* summed square error originla/collage */
      best_se,          
      best_se_vq,
      best_r_alpha,
      se_vq,
      sum=0,
      sum2=0,
      beta,                   
      r_alpha, 
      r_beta,
      obj_val,
      best_obj_val;
  double *phi_r, *phi_neg_r, *phi_pos_r;     /* phi of range           */
  struct classified_domain *domain_node;     /* var for domains we scan through */
  
  n = depth -min_part;
  phi_r=(double *)malloc(blocksize * blocksize * sizeof(double));
  
  best_se = HUGE;      /* just a big number */
  best_se_vq =HUGE;
  
  classify(atx, aty, blocksize, blocksize,
	   &the_first_class,
	   &phi_r,
	   range_sym_op, &sum, &sum2, 1);
  
  if (phi_r != NULL) /* original  imageblock has norm != 0 */
    {
      /* We must CHECK EACH DOMAIN TWICE. Once for positive scaling,  */
      /* once for negative scaling. Each has its own class and sym_op */
      if (!only_positive) 
	  {
	    range_sym_op[1] = 
	      rot_transform[(the_first_class == 2 ? 1 : 0)][range_sym_op[0]];
	  } 
      
      for(i=0; i<nn[n]; ++i)
	list_dist[i] = HUGE;
      
      phi_pos_r = (double *)malloc(dim * sizeof(double));
	
      for(j=0;j<dim;j++)
	phi_pos_r[j] = phi_r[phi_trafo[range_sym_op[0]][j]];
      
      cluster_search(phi_pos_r, list_clu, list_dist, list_sgn,
		    1, nn[n], n);
      
      free(phi_pos_r);
      
      m=0;
      if(!best)
	{
	  m=nn[n]>>1;
	  for(i=m;i<nn[n];++i)
	    list_dist[i]=HUGE;
	}
      
      /* only_positive is 0 or 1, so we may or not scan */
      if(!only_positive)
	{ 
	  phi_neg_r = (double *)malloc(dim * sizeof(double));
	  for(j=0;j<dim;j++)
	    phi_neg_r[j] = phi_r[phi_trafo[range_sym_op[1]][j]];
	  cluster_search(phi_neg_r, &list_clu[m], &list_dist[m], &list_sgn[m],
		         -1, nn[n]-m, n);
	  free(phi_neg_r);
	}
      free(phi_r);
      
      /**********************************************************/
      /**                 F R A C T A L                         */
      /**/
      for(j=0; j<nn[n]; j++)
	{
	  i=list_sgn[j];
	  for (domain_node = the_domain[list_clu[j]][n];
	       domain_node != NULL; domain_node = domain_node->next) 
	    {
	      domain_sym_op = domain_node->the->sym;

	      /* The following if statement figures out how to transform  */
	      /* the domain onto the range, given that we know how to get */
	      /* each into cannonical position.                           */
	      if (((domain_sym_op>3 ? 4: 0) + (range_sym_op[i]>3 ? 4: 0))%8 == 0)
		sym_op = (4 + domain_sym_op%4 - range_sym_op[i]%4)%4;
	      else 
		sym_op = (4 + (domain_sym_op%4 + 3*(4-range_sym_op[i]%4))%4)%8;
	      
	      se = compare_fr(atx,aty, blocksize, blocksize,  depth, sum,sum2, 
			    domain_node->the->dom_x, 
			    domain_node->the->dom_y, 
			    sym_op, &ialpha,&ibeta); 
	      
	      if (se < best_se) 
		{    
		  best_ialpha = ialpha;   
		  best_ibeta = ibeta;
		  best_se = se;
		  best_sym_op = sym_op;
		  best_domain_x = domain_node->the->dom_x;
		  best_domain_y = domain_node->the->dom_y;
		}
	    }
	}
      
      /* compute the objective value  dist + lambda * rate
         rate = ... + 5 with 5=3 bits for isometry operation plus 2 bits for 
         the used method, vq or fractal */
      objective_val[0] = lambda * (s_bits+o_bits+5+bits_needed[depth-min_part])
	                 + best_se;
      
      /**********************************************************/
      /**                ORMSGVQ                        */
      /**/
      l = (blocksize == 4) ? 0 : ( (blocksize == 8) ? 1 : 2 ); 
      best_obj_val = HUGE;
      
      if(blocksize  < 17)
	{
	  domain_sym_op = 0;
	 	 	  
	  for(j=0; j<nn[n]; j++)
	    {
	      i=list_sgn[j];
	      if (((domain_sym_op>3 ? 4: 0) + (range_sym_op[i]>3 ? 4: 0))%8 == 0)
		sym_op = (4 + domain_sym_op%4 - range_sym_op[i]%4)%4;
	      else
		sym_op = (4 + (domain_sym_op%4 + 3*(4-range_sym_op[i]%4))%4)%8;
	      
	      se_vq = compare_cen(atx,aty, blocksize, blocksize, depth, sum,sum2, 
				  list_clu[j],sym_op, &ialpha, &ibeta,
				  &r_alpha, &r_beta);      
	      obj_val = se_vq + lambda * codelength[l][ialpha];
	      
	      if (obj_val < best_obj_val)
		{
		  best_ialpha_vq = ialpha;      
		  best_ibeta_vq = ibeta;
		  best_obj_val = obj_val;
		  best_sym_op_vq = sym_op;
		  best_se_vq = se_vq;
		  best_vec = list_clu[j];
		  best_r_alpha = r_alpha; 
		}
	    }
	}
      objective_val[1] = best_obj_val + lambda * (4+vec_bits+o_bits);
    }

  /**********************************************************/
  /**               C O N S T A N T                         */
  /**/  
  beta=sum/(blocksize*blocksize);
  best_ibeta_c = 0.5 + beta/GREY_LEVELS*((1<<o_bits)-1);
  beta = (double)best_ibeta_c/(double)((1<<o_bits)-1)*GREY_LEVELS;
  se_vq = sum2 + beta*(beta*blocksize*blocksize - 2.0*sum);
  objective_val[2] = se_vq + lambda * (o_bits+2);
  
  /* FIND coding scheme with MINIMAL OBJECTIVE VALUE */
  best_obj_val=HUGE;
  for (i=0; i<3; i++)
    {
      if(objective_val[i] < best_obj_val)
	{
	  best_method = i;
	  best_obj_val = objective_val[i];
	}
    }
  
  /* STORE the correct optimal DATA in quadtree node */
  if (best_method == 0)   
    { /* fractal coding */
      node->method = 0;
      node->iso = best_sym_op;
      node->scal = best_ialpha;
      node->offset = best_ibeta;
      node->block_x = best_domain_x;
      node->block_y = best_domain_y;
      node->se = best_se;
      node->rate = s_bits+o_bits+5+bits_needed[depth-min_part];
    }
  if (best_method == 1)   
    { /* OMRSG-VQ */         
      node->method = 1;
      node->iso = best_sym_op_vq;
      node->scal = best_ialpha_vq;
      node->offset = best_ibeta_vq;
      node->block_x = best_vec;
      node->se = best_se_vq;
      node->rate = codelength[l][best_ialpha_vq]+o_bits+4+vec_bits;;
    }
  if (best_method == 2) 
    { /* constant: s=0 */
      node->method = 2;
      node->offset = best_ibeta_c; 
      node->se = se_vq;
      node->rate = o_bits+2;
    }

  /* RECURSIVELY TRAVERSE QUADTREE one layer deeper if possible */ 
  if (depth < max_part) 
    {
      node->quadtree = 1;
      encode_quadtree(atx,aty,blocksize/2,depth+1,lambda,node->son1);
      encode_quadtree(atx+blocksize/2,aty,blocksize/2,depth+1,lambda,node->son2);
      encode_quadtree(atx,aty+blocksize/2,blocksize/2,depth+1,lambda,node->son3);
      encode_quadtree(atx+blocksize/2,aty+blocksize/2,blocksize/2,
		      depth+1,lambda,node->son4);
    }
  else {
    node->quadtree = 0;
    return;
  }
}


    
  
