/******************************************************************************
 *
 * NAME
 *    encode.functions.c (used by encode.c)
 *    
 *    see encode.h for definitions of default values
 * 
 * VERSION
 *    1.0
 *     
 * DATE
 *    10.9.1998
 *
 * IMPLEMENTATION 
 *    Bertram Ganz (ganz@informatik.uni-freiburg.de)
 *    Martin Mueller (mueller@www.fr.bw.schule.de)
 *    
 *    The fractal part uses Yuval Fisher's quadtree code [2] 
 *	
 * 
 **************************************************************************/

#include "encode.h"

/****************************************************************/
/* ARRAY DECLARATIONS and INITIALIZATIONS                     */
/****************************************************************/
const double e[16]            = {0.25,0.25,0.25,0.25,0.25,0.25,0.25,0.25,\
				 0.25, 0.25,0.25,0.25,0.25,0.25,0.25,0.25};    
const int rot_transform[2][8] = {7,4,5,6,1,2,3,0, 2,3,0,1,6,7,4,5};
const int phi_trafo[8][16]    = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
				 3,7,11,15,2,6,10,14,1,5,9,13,0,4,8,12,
				 15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0,
				 12,8,4,0,13,9,5,1,14,10,6,2,15,11,7,3,
				 0,4,8,12,1,5,9,13,2,6,10,14,3,7,11,15,
				 3,2,1,0,7,6,5,4,11,10,9,8,15,14,13,12,
				 15,11,7,3,14,10,6,2,13,9,5,1,12,8,4,0,
				 12,13,14,15,8,9,10,11,4,5,6,7,0,1,2,3};
const int codelength[3][32]   = {6,8,7,7,6,6,6,5,5,5,5,4,4,4,4,6,    /* 4x4 */ 
                                 7,4,4,4,4,4,4,5,5,6,6,6,7,7,8,6,
			     	 8,8,8,8,8,7,6,6,5,5,4,4,4,4,4,5,    /* 8x8 */
				 7,4,3,4,4,4,5,5,5,6,6,7,7,8,8,8, 
				 10,10,10,10,8,7,7,7,6,6,5,4,4,3,3,  /* 16x16 */  
				 5,5,3,3,4,4,5,6,6,7,7,8,8,8,9,10,10};
                                 /* this array contains the codeword lengths */
                                 /* of our separately generated Huffmancode */
                                 /* being used for coding the VQ scaling factors .*/
const char huff_codeword[3][32][13] =  /* 16x16 */ 
                                {"0011011010","1101010101","1101010100", 
				 "0011011000","10100100","1101011","1010001",
				 "1010000","101011","101010","11011","1011",
				 "1001","111","011","10000","10001","010",
				 "000","0010","1100","00111","001100","110100",
				 "1010011","0011010","00110111","10100101",
				 "11010100","110101011","0011011011",
				 "0011011001",
				 /* 8x8 */
				 "10011110","10011111","01101100","00110110", 
				 "00110101","0001101","011010","001100","10010",
				 "00111","1010","1111","1011","1110","0101",
				 "10001","1001110","0100","110","0000","0010",
				 "0111","00010","01100","10000","000111",
				 "100110","0001100","0110111","00110100",
				 "00110111","01101101",
				 /* 4x4 */  
				 "111000","00011111","1111111","0011101", 
				 "111110","111010","000001","11110","00001",
				 "00101","00010","1100","1010","0100","0111",
				 "001111","1111110","0110","0101","1001",
				 "1011","1000","1101","00100","00110","000000",
				 "000110","111011","0001110","0011100",
				 "00011110","111001"};

/***************************************************************/
/* FATAL is for when there is a fatal error... print a message */ 
/* and exit.                                                   */
/***************************************************************/
void fatal(char *s)
{
     printf("%s\n",s);
     exit(-1);
}

/***************************************************************/
/* INITIALIZATION of global variables                          */
/***************************************************************/
void initialize()
{
  max_scale_vq[0]  = 250;
  max_scale_vq[1]  = 500;
  max_scale_vq[2]  = 1000;
  max_scale        = 1.0;

  counter.vq       = 0; 
  counter.fr       = 0; 
  counter.mean     = 0; 

  s_bits           = 5;           
  o_bits           = 7;           
  min_part         = 4;         
  max_part         = 6;         
  dom_step         = 1;         
  dom_step_type    = 0;    
  dom_type         = 0;         
  only_positive    = 0;    
  subclass_search  = 0;  
  fullclass_search = 0; 
  dim              = 16;             
  
  best             = 1;             
  
  number_nn_class  = 1;  
  c_x              = 16;             
  c_y              = 16;		     
}

/****************************************************************/
/* READ IN command line arguments                                */
/****************************************************************/
void read_commandline(int *argc, char **argv, double *lambda, 
		      char *inputfilename, char *outputfilename, 
		      int *hsize, int *vsize, long *stripchar)
{
  int i;

  for (i=1; i<*argc; ++i) 
    if (argv[i][0] != '-' ) 
      if (inputfilename[0] == 1)
	strcpy(inputfilename, argv[i]);
      else if (outputfilename[0] == 1)
	strcpy(outputfilename, argv[i]);
      else;
    else 
      { /* we have a FLAG */
	if (strlen(argv[i]) == 1) break;
	switch(argv[i][1]) {
	case 'l': *lambda = atof(argv[++i]);
	  break;
	case 'S': *stripchar = atoi(argv[++i]);
	  break;
	case 'x': c_x = atoi(argv[++i]);
	  break;
	case 'y': c_y = atoi(argv[++i]);
	  break;
	case 'z': best = atoi(argv[++i]);
	  break;
	case 'Z': number_nn_class = atoi(argv[++i]);
	  break;
	case 'w': *hsize = atoi(argv[++i]);
	  break;
	case 'h': *vsize = atoi(argv[++i]);
	  break;
	case 'D': dom_type = atoi(argv[++i]);
	  break;
	case 'd': dom_step = atoi(argv[++i]);
	  if (dom_step < 0 || dom_step > 15) 
	    fatal("\n Bad domain step.");
	  break;
	case 's': s_bits = atoi(argv[++i]);
	  break;
	case 'o': o_bits = atoi(argv[++i]);
	  break;
	case 'm': min_part = atoi(argv[++i]);
	  break;
	case 'M': max_part = atoi(argv[++i]);
	  break;
	case 'e': dom_step_type= 1;
	  break;
	case 'p': only_positive = 1;
	  break;
	case 'N': max_scale = atof(argv[++i]);
	  break;
	case 'V': if (strlen(argv[i]) == 2) fatal("bad parameter");
	  switch(argv[i][2]) 
	    {
	    case '4': max_scale_vq[0] = atof(argv[++i]);
	      break;
	    case '8': max_scale_vq[1] = atof(argv[++i]);
	      break;
	    case '9': max_scale_vq[2] = atof(argv[++i]);
	      break;
	    default: break;
	    }; break;
	case '?':
	case 'H':
	default:
	  printf("\nUsage: encode -[options] [inputfile [outputfile]]");
	  printf("\nOptions are: (# = number), defaults show in ()");
	  printf("\n  -l # Lagrange multiplier. (%lf)", *lambda);
	  printf("\n  -m # minimum quadtree partitions. (%d)",min_part);
	  printf("\n  -M # maximum quadtree partitions. (%d)",max_part);
	  printf("\n  -S # number of input bytes to ignore. (%ld)",stripchar);
	  printf("\n  -w # width (horizontal size) of input data. (512)");
	  printf("\n  -h # height (vertical size) of input data. (512)");
	  printf("\n  -x # horizontal number of cluster-centers. (%d)", c_x);
	  printf("\n  -y # vertical number of cluster-centers. (%d)", c_y);
	  printf("\n  -Z # number of centers beeing visited for nn-search (%d)",
		 best);
	  printf("\n  -z # best of positive/negative phi's (on)(%d)",
		 number_nn_class);
	  printf("\n  -d # domain step size. (%d)", dom_step);
	  printf("\n  -D # method {0,1,2} for domain pool (%d)",dom_type);
	  printf("\n  -s # number of scaling quantizing bits. (%d)",s_bits);
	  printf("\n  -o # number of offset quantizing bits. (%d)",o_bits);
	  printf("\n  -N # maximum scaling in encoding. (%lf)",max_scale);
	  printf("\n  -V4 # maximum scaling in VQ encoding (4x4). (%lf)",
		 max_scale_vq[0]);
	  printf("\n  -V8 # maximum scaling in VQ encoding (8x8). (%lf)",
		 max_scale_vq[1]);
	  printf("\n  -V9 # maximum scaling in VQ encoding (16x16). (%lf)",
		 max_scale_vq[2]);
	  printf("\n  -e   domain step as multiplier not divisor. (off)");
	  printf("\n  -p   use only positive scaling (for speed). (off)");
	  fatal("\n        --------- Version 1.0 --------------");
	}
      }
}

/****************************************************************/
/* FREE MEMORY allocated in the list structure the_domain       */
/****************************************************************/
void list_free(struct classified_domain *node)
{
     if (node->next != NULL)
       list_free(node->next);
     free(node);
}

/***************************************************************/
/* return the AVERAGE PIXEL VALUE of a region of the image.    */
/***************************************************************/
void average(int x, int y, int xsize, int ysize, double *psum, double *psum2)
{
    register int i,j,k;
    register double pixel;
    *psum = *psum2 = 0.0;
    k = ((x%2)<<1) + y%2;
    x >>= 1; y >>= 1;
    xsize >>= 1; ysize >>= 1;
    for (i=x; i<x+xsize; ++i)
    for (j=y; j<y+ysize; ++j) {
	   pixel = domimage[k][j][i];
       *psum += pixel;
       *psum2 += pixel*pixel;
    }
}

/******************************************************************/
/* return the AVERAGE PIXEL VALUE of a region of the image. This  */
/* routine differs from the previous in one slight way. It does   */
/* not average 2x2 sub-images to pixels. This is needed for clas- */
/* sifying ranges rather than domain where decimation is needed.  */
/******************************************************************/
void average1(int x, int y, int xsize, int ysize, double *psum, double *psum2)
{
    register int i,j;
    register double pixel;
    *psum = *psum2 = 0.0;

    for (i=x; i<x+xsize; ++i)
    for (j=y; j<y+ysize; ++j) {
	   pixel = (double)image[j][i];
       *psum += pixel;
       *psum2 += pixel*pixel;
    }
}

/*********************************************************/
/* COMPUTE NORM of a                                     */
/*********************************************************/
double norm(double *a, int dim)  
{
  int i;
  double s=0.;
  for(i=0; i<dim; i++)
    s+=a[i]*a[i];
  return s;
}

/**********************************************************/
/* COMPUTE PROJECTION phi of x                            */
/**********************************************************/
void compute_phi(double *x, double **phi, int dim, double x_sum)
{
  int i;
  double s, *z, sum;

  z=(double *)malloc(dim * sizeof(double));
  sum = x_sum * 0.0625;
  for(i=0; i<dim; i++)
    z[i]=x[i]-sum;
  s = sqrt(norm(z,dim));
  if(s!=0)               
    { 
      for(i=0; i<dim; i++)
	(*phi)[i] = z[i]/s;  
      free(z);
    }
  else /* z is parallel to e      */
    {  /* we don't need phi later */
      free(*phi);
      *phi=NULL;
    }                    
}

/**********************************************************/
/* COMPUTE  SQUARED EUCLIDEAN DISTANCE between a and b.        */
/* sgn is a variable for negative phi.                    */
/**********************************************************/
double dist(double *a, double *b, signed int sgn) 
{
  int    i;
  double sum=0.;
  
  for(i=0;i<16;i++)
    sum += (a[i]-(sgn*b[i]))*(a[i]-(sgn*b[i]));
  
  return sum;
}

/*********************************************************/
/* COMPUTE DISTANCE for normed vectors                   */
/*********************************************************/
double distnorm(double *a, double *b, signed int sgn)
{
  int    i;
  double sum=0.;

  for(i=0;i<16;i++)
    sum += (a[i]*b[i]);

  return 2-2*sgn*sum;
}

/**********************************************************/
/* NN_CLUSTER_SEARCH searches for the nearest cluster     */
/* center and returns the square error and the correspon- */
/* ding cluster number.                                   */
/**********************************************************/
double nn_cluster_search(double *phi, int *cl_nr, signed int sgn, int depth)
{
  double d, dist_ = HUGE;
  int    m;

  for(m=0; m<(size_clus); m++)
    {
      if ((d = distnorm(centers[depth][m], phi, sgn)) < dist_)
        {
          *cl_nr = m;
          dist_ = d;
        }
    }
  return dist_;
}  

/**************************************************************/
/* CLUSTER_SEARCH searches for the  number nearest neighbors. */
/**************************************************************/
void cluster_search(double *phi, int *cl_nr, double *list, int *l_sgn,
		 signed int sgn, int last, int depth)
{
  double d;
  int    i, m;

  for(m=0; m<dim_clus[depth]; m++)
    {
      i = last - 2;
      d = distnorm(centers[depth][m], phi, sgn);
      if (d<list[last-1]) /* last: index of last element in list */
	{
	  list[last-1]=d; cl_nr[last-1]=m; 
	  if(sgn==1)l_sgn[last-1]=0; 
	  else l_sgn[last-1]=1;
	}
      while((i>-1) && (list[i+1]<list[i]))
	{
	  swap(list[i+1], list[i], double);
	  swap(cl_nr[i+1], cl_nr[i], int);
	  swap(l_sgn[i+1], l_sgn[i], int);	  
	  i--;
	}
    }
}  

/**************************************************************/
/* PACK VALUE using size bits and output into foutf           */
/* which_file specifies the file to be packed                 */
/* which_file 0: main data_file/1: method_file/2 offset_file  */
/**************************************************************/
int pack(int size, long int value, int which_file, FILE *foutf)
{
     int i;
     static int ptr[3]={1,1,1}, /* how many bits are packed in sum so far */
                sum[3]={0,0,0}, /* packed bits */ 
                num_of_packed_bytes[3]={0,0,0}; /* total bytes written out */

    /* size == -1 means we are at the end, so write out what is left */
    if (size == -1 && ptr[which_file] != 1) {
      fputc(sum[which_file]<<(8-ptr[which_file]), foutf);
      ++num_of_packed_bytes[which_file];
      return(0);
    }
    
    /* size == -2 means we want to know how many bytes we have written */
    if (size == -2) 
      return(num_of_packed_bytes[which_file]);
    
    for (i=0; i<size; ++i, ++ptr[which_file], 
	 value = value>>1, sum[which_file] = sum[which_file]<<1) {
      if (value & 1) 
	sum[which_file] |= 1;
      if (ptr[which_file] == 8) {
	fputc(sum[which_file],foutf);
	++num_of_packed_bytes[which_file];
	sum[which_file]=0;
	ptr[which_file]=0;
      }
    }
}

/************************************************************/
/* QUADTREE_ALLOC recursively allocates memory for quadtree */
/* of depth 'depth'.                                        */ 
/************************************************************/
void quadtree_alloc(struct quadtree_node *node, int depth)
{
  int i;
  struct quadtree_node *new_son[4];
  
  if (depth>1) {
    for (i=0; i<4; i++) 
      new_son[i] = (struct quadtree_node*) 
	           malloc(sizeof(struct quadtree_node));
    
    node->son1 = new_son[0];  node->son2 = new_son[1];  
    node->son3 = new_son[2];  node->son4 = new_son[3]; 
    
    for (i=0; i<4; i++) 
      quadtree_alloc(new_son[i],depth-1);
    }
  return;
}


/********************************************************************/
/* VISIT_LEAVES descends quadtree from given node and sums up all */
/* relevant data (rate, summed square error) of final leaves.       */
/********************************************************************/
void visit_leaves(struct quadtree_node *node, double *se_leaves,
		  int *rate_leaves, int *qt_bits, int depth)
{
  if (node->quadtree == 0)
    {
      *se_leaves += node->se;
      *rate_leaves += node->rate;
      if (depth==max_part)
	*qt_bits -= 1;
      return;
    }
  else 
    {
      *qt_bits +=4;
      visit_leaves(node->son1,se_leaves,rate_leaves,qt_bits,depth+1);
      visit_leaves(node->son2,se_leaves,rate_leaves,qt_bits,depth+1);
      visit_leaves(node->son3,se_leaves,rate_leaves,qt_bits,depth+1);
      visit_leaves(node->son4,se_leaves,rate_leaves,qt_bits,depth+1);
    }
}

/**********************************************************/
/* PRUNE_QUADTREE prunes the quadtree using the criterion */
/* delta_dist/delta_rate <=lambda.                         */
/**********************************************************/
void prune_quadtree(double lambda, struct quadtree_node *node, 
		    int depth, int prune_depth)
{
  double se_leaves=0, 
         delta_se;
  int rate_leaves=0, 
      delta_rate,
      qt_bits=0;  
    
  if (depth != prune_depth)
    {
      prune_quadtree(lambda,node->son1,depth+1,prune_depth);
      prune_quadtree(lambda,node->son2,depth+1,prune_depth);
      prune_quadtree(lambda,node->son3,depth+1,prune_depth);
      prune_quadtree(lambda,node->son4,depth+1,prune_depth);
    }
  else 
    {
      visit_leaves(node,&se_leaves,&rate_leaves,&qt_bits,prune_depth);
      delta_rate = rate_leaves + qt_bits - node->rate;
      delta_se   = node->se - se_leaves;
      
      if( (delta_se/(double)delta_rate)<=lambda)
	node->quadtree = 0;
    }
}     

/********************************************************************/
/* STORE_QTDATA stores the complete quadtree data in code[k][i][j]. */
/********************************************************************/
void store_qtdata(int atx, int aty, struct quadtree_node *node, 
		  int depth, int blocksize)
{
  int i,j,k;

  i=atx/blocksize; j=aty/blocksize;
  k=depth-min_part;

  code[k][i][j].method   = node->method;
  code[k][i][j].quadtree = node->quadtree;
  code[k][i][j].block_x  = node->block_x;
  code[k][i][j].block_y  = node->block_y;
  code[k][i][j].scal     = node->scal;
  code[k][i][j].offset   = node->offset;
  code[k][i][j].iso      = node->iso;

  if (depth < max_part) {
    store_qtdata(atx,aty,node->son1,depth+1,blocksize/2);
    store_qtdata(atx+blocksize/2,aty,node->son2,depth+1,blocksize/2);
    store_qtdata(atx,aty+blocksize/2,node->son3,depth+1,blocksize/2);
    store_qtdata(atx+blocksize/2,aty+blocksize/2,node->son4,depth+1,blocksize/2);
  }
  else
    return;
}

/********************************************************************/
/* SAVE_DATA passes the quadtree to save the main data in  */
/* outputfile. Main data is information about isometries, domain-   */
/* or codebook numbers, quadtree and scaling factors.               */
/********************************************************************/
void save_data(int atx, int aty, int blocksize, int depth)
{
  int i,j,k,l,m,bit;

  /* KEEP BREAKING it down until we are small enough */
  if (depth<min_part) {
    save_data(atx,aty,blocksize/2,depth+1);
    save_data(atx+blocksize/2,aty,blocksize/2,depth+1);
    save_data(atx,aty+blocksize/2,blocksize/2,depth+1);
    save_data(atx+blocksize/2,aty+blocksize/2,blocksize/2,depth+1);
    return;
  }

  i = atx/blocksize; j = aty/blocksize;
  k = depth-min_part;
  if (code[k][i][j].quadtree == 1 && depth < max_part) 
    {
      pack(1,(long)1,2,outputfile); /* This bit means we quadtree'd down */
      save_data(atx,aty,blocksize/2,depth+1);
      save_data(atx+blocksize/2,aty,blocksize/2,depth+1);
      save_data(atx,aty+blocksize/2,blocksize/2,depth+1);
      save_data(atx+blocksize/2,aty+blocksize/2,blocksize/2,depth+1);
    }
  else 
    {
      /* The fit was good enough or we just can't get smaller ranges */
      /* So WRITE OUT THE DATA */
      if (depth < max_part)           /* if we are not at the smallest range */
	pack(1,(long)0,2,outputfile); /* then we must tell the decoder we    */
                                      /* quadtreed down */
      
       switch(code[k][i][j].method) {
       case 2: ++counter.mean; /* Use constant */
	 break;
       case 1: ++counter.vq;   /* OMSGVQ if not constant */
	 m=(blocksize == 16) ? 0 : ( (blocksize == 8) ? 1 : 2 ); 
	 l=0; /* PACK VARIABLE LENGTH CODE (huffman) for scal.-factor */
	 while (huff_codeword[m][code[k][i][j].scal][l]!='\0')
	   {
	     bit = (huff_codeword[m][code[k][i][j].scal][l++]=='0')?0:1;
	     pack(1, (long)bit,2, outputfile); 
	   }
	 pack(3, (long)code[k][i][j].iso,2, outputfile);
	 pack(vec_bits, (long)code[k][i][j].block_x,2, outputfile);
	 break;
       case 0: ++counter.fr;   /* Use fractal scheme */
	 pack(s_bits, (long)code[k][i][j].scal,2, outputfile); 
	 pack(3, (long)code[k][i][j].iso,2, outputfile);
	 pack(bits_needed[depth-min_part], (long)(code[k][i][j].block_y*
	      domain.no_h_domains[depth-min_part]+code[k][i][j].block_x),2,
	      outputfile);
	 break;
       }
    }
}    


/***********************************************************************/
/* SAVE_DATA passes the quadtree to save information about    */
/* used method and offset in additional files (methodfile, offsetfile) */
/***********************************************************************/
void save_data1(int atx, int aty, int blocksize, int depth)
{
  int i,j,k;
  
  /* KEEP BREAKING IT DOWN until we are small enough */
  if (depth<min_part) 
    {
      save_data1(atx,aty,blocksize/2,depth+1);
      save_data1(atx+blocksize/2,aty,blocksize/2,depth+1);
      save_data1(atx,aty+blocksize/2,blocksize/2,depth+1);
      save_data1(atx+blocksize/2,aty+blocksize/2,blocksize/2,depth+1);
      return;
    }
  
  i = atx/blocksize; j = aty/blocksize;
  k = depth-min_part;
  
  if (code[k][i][j].quadtree == 1 && depth < max_part) 
    {
      save_data1(atx,aty,blocksize/2,depth+1);
      save_data1(atx+blocksize/2,aty,blocksize/2,depth+1);
      save_data1(atx,aty+blocksize/2,blocksize/2,depth+1);
      save_data1(atx+blocksize/2,aty+blocksize/2,blocksize/2,depth+1);
    }
  else 
    {
      switch(code[k][i][j].method) {
      case 2: /* Use constant */
	pack(1,(long)0,0,methodfile);  /* 00: s=0-info for decoder */
	pack(1,(long)0,0,methodfile);
	pack(8, (long)code[k][i][j].offset,1, offsetfile); 
	break;
      case 1: /* Use OMRSG-VQ */ 
	pack(1,(long)1,0,methodfile);  /* 1: VQ,s|=0-info for decoder */
	pack(8, (long)code[k][i][j].offset,1, offsetfile); 
	break;
      case 0: /* Use Fractal */ 
	pack(1,(long)0,0,methodfile);  /* 01: use fractal info for decoder */
	pack(1,(long)1,0,methodfile); 
	pack(8, (long)code[k][i][j].offset,1, offsetfile); 
	break;
      }
    }
}

/***********************************************************************/
/* MERGE_DATE calls SAVE_DATA-routines an checks wether gzip-compres-  */
/* sion of method and offset data is useful or not. The final output- */
/* file contains header information, a block of method bits, a block   */
/* of offset bits and a block of main data.                            */
/***********************************************************************/
void merge_outputdata(char *outputfilename, int hsize, int vsize)
{
  struct stat  methodfile_data, offsetfile_data;
  int          methodfile_bytes_before, methodfile_bytes_after,
               methodfile_bytes,
               offsetfile_bytes_before, offsetfile_bytes_after,
               offsetfile_bytes,
               outputfile_bytes;
  double       comp;
  char         *data;
  FILE         *inputfile;
  
  if ((outputfile = fopen(outputfilename, "w")) == NULL)
    fatal("Can't open output file.");

  /* OUTPUT FIRST PART OF HEADER-DATA into the outputfile.*/
  pack(4,(long)min_part,2,outputfile);
  pack(4,(long)max_part,2,outputfile);
  pack(4,(long)dom_step,2,outputfile);
  pack(1,(long)dom_step_type,2,outputfile);
  pack(2,(long)dom_type,2,outputfile);
  pack(12,(long)hsize,2,outputfile);
  pack(12,(long)vsize,2,outputfile);
  pack(4, (long)vec_bits,2,outputfile);
  
  /* SYSTEMCALL FOR GZIP compression after having saved method- 
     and offset-information */
  if ((methodfile = fopen("methodfile", "w")) == NULL)
    fatal("Can't open methodfile.");
  if ((offsetfile = fopen("offsetfile", "w")) == NULL)
    fatal("Can't open offsetfile.");
  
  /* SAVE SIDE-INFORMATION-DATA (offset, method) in separate files */
  save_data1(0,0,hsize,0);
  pack(-1,(long)0,0,methodfile);
  pack(-1,(long)0,1,offsetfile);
  fclose(methodfile);
  fclose(offsetfile);
  
  stat("methodfile", &methodfile_data);
  stat("offsetfile", &offsetfile_data);
  
  methodfile_bytes_before = methodfile_data.st_size;
  offsetfile_bytes_before = offsetfile_data.st_size;
  
  system("gzip -9 -f methodfile > /dev/null");
  system("gzip -9 -f offsetfile > /dev/null");

  stat("methodfile.gz", &methodfile_data);
  stat("offsetfile.gz", &offsetfile_data);
  
  methodfile_bytes_after = methodfile_data.st_size;
  offsetfile_bytes_after = offsetfile_data.st_size;

  if(methodfile_bytes_after<methodfile_bytes_before)
    {
      pack(1,(long)1,2,outputfile); /* gzip is used */
      pack(12,(long)methodfile_bytes_after,2,outputfile);
      data = (char *)malloc(sizeof(char)*methodfile_bytes_after);
      if ((inputfile = fopen("methodfile.gz", "r")) == NULL)
	fatal("Can't open methodfile.gz");
      fread(data,sizeof(char),methodfile_bytes_after,inputfile); 
      fclose(inputfile);
      fwrite(data,sizeof(char),methodfile_bytes_after,outputfile);
      free(data);
      system("rm methodfile.gz");
      methodfile_bytes = methodfile_bytes_after;
    }
  else 
    {
      pack(1,(long)0,2,outputfile); /* gzip is not used */
      pack(12,(long)methodfile_bytes_before,2,outputfile);
      system("gzip -d -9 -f methodfile.gz > /dev/null");
      data = (char *)malloc(sizeof(char)*methodfile_bytes_before);
      if ((inputfile = fopen("methodfile", "r")) == NULL)
	fatal("Can't open methodfile");
      fread(data,sizeof(char),methodfile_bytes_before,inputfile); 
      fclose(inputfile);
      fwrite(data,sizeof(char),methodfile_bytes_before,outputfile);
      free(data);
      system("rm methodfile");
      methodfile_bytes = methodfile_bytes_before;
    }
  
  if(offsetfile_bytes_after<offsetfile_bytes_before)
    {
      pack(1,(long)1,2,outputfile); /* gzip is used */
      pack(15,(long)offsetfile_bytes_after,2,outputfile);
      data = (char *)malloc(sizeof(char)*offsetfile_bytes_after);
      if ((inputfile = fopen("offsetfile.gz", "r")) == NULL)
	fatal("Can't open offsetfile.gz");
      fread(data,sizeof(char),offsetfile_bytes_after,inputfile); 
      fclose(inputfile);
      fwrite(data,sizeof(char),offsetfile_bytes_after,outputfile);
      free(data);
      system("rm offsetfile.gz");
      offsetfile_bytes = offsetfile_bytes_after;
    }
  else 
    {
      pack(1,(long)0,2,outputfile); /* gzip is not used */
      pack(15,(long)offsetfile_bytes_before,2,outputfile);
      system("gzip -d -9 -f offsetfile.gz > /dev/null"); /* decompress offsets */
      data = (char *)malloc(sizeof(char)*offsetfile_bytes_before);
      if ((inputfile = fopen("offsetfile", "r")) == NULL)
	fatal("Can't open offsetfile");
      fread(data,sizeof(char),offsetfile_bytes_before,inputfile); 
      fclose(inputfile);
      fwrite(data,sizeof(char),offsetfile_bytes_before,outputfile);
      free(data);
      system("rm offsetfile");
      offsetfile_bytes = offsetfile_bytes_before;
    }
  
  /* now we have to PACK ALL THE REST (main data) in outputfile */
  save_data(0,0,hsize,0);
  
  /* STUFF THE LAST BYTE if needed */
  pack(-1,(long)0,2,outputfile);
  fclose(outputfile);
  
  printf("\nDone.\n\n");
  printf("Number of ranges coded with cluster centers: %i\n", counter.vq );
  printf("Number of ranges coded with domains: %i\n", counter.fr );
  printf("Number of ranges coded with their mean: %i\n", counter.mean);
  fflush(stdout);

  outputfile_bytes = pack(-2,(long)0,2,outputfile)+methodfile_bytes
                     +offsetfile_bytes;
  comp = (double)(hsize*vsize)/(double) outputfile_bytes;
  printf("\nCompression = %lf from %d bytes written in %s.\n",
	 comp, outputfile_bytes, outputfilename);
}
